/*
 *
 * B2BUACall.cxx
 * 
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: B2BUACall.cxx,v $
 * Revision 1.71  2007/07/26 11:52:56  joegenbaclor
 * Finalized AuthenticationPending
 *
 * Revision 1.70  2007/07/16 11:41:59  joegenbaclor
 * Flagged ICT final response retransmissions.  This resulted to calling SendB2BConnect multiple times.
 *
 * Revision 1.69  2007/07/15 12:33:19  joegenbaclor
 * Corrected bug in calling OnIncomingSDPOffer where it is called for the second leg instead of the first leg call session.  This is due to the change in placement of the method form receipt of the invite to SetupOutbound
 *
 * Revision 1.68  2007/07/15 07:50:04  joegenbaclor
 * tidying up resource counters
 *
 * Revision 1.67  2007/07/14 07:50:29  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.66  2007/07/06 13:38:39  rcolobong
 * Add Upstream Proxy Route feature
 *
 * Revision 1.65  2007/06/30 04:17:28  joegenbaclor
 * Moved processing of SDPOffer from B2BUACall::IST_OnReceivedInvite to B2BUACall::OnSetupOutbound
 *
 * Revision 1.64  2007/06/26 02:22:53  joegenbaclor
 * Bug fix:  SIPHeaderB operator=() does not call MakeUnique on parameters.   The effect
 *  of this is when SIPSessions send responses with a to tag, the original INVITE would
 *  inherit the tag automatically.   This might be a desired behavior but might lead
 *  to confusion as how the original INVITE got the tag.
 *
 * B2BUACall And CallSession are updated accordingly to comply with this fix
 *
 * Revision 1.63  2007/06/19 11:25:13  joegenbaclor
 * Allowed Subscribe, Notify and Publish to be proxied by the B2BUA Endpoint.  We did this by introducing custom ProcessStackEvent handlers for B2BUAEndPoint
 *
 * Revision 1.62  2007/06/16 12:30:40  joegenbaclor
 * Changed CacheManager to return a smart pointer instead of a direct pointer to cache object to avoid segmentation faults when an object expires in cache and another thread has pending ownership of the pointer
 *
 * Renamed CacheManager::GetAt() to CacheManager::CreatePointer() to give a hint to users that the pointer needs to be deleted after being used
 *
 * Revision 1.61  2007/06/07 12:22:00  rcolobong
 * 1. Bug fix errors related to Consultative Transfer
 * 2. Update handling Replaces during Refer
 * 3. Add new method ConstructReplaces in SIPSession
 *
 * Revision 1.60  2007/05/25 08:38:09  joegenbaclor
 * Introduced GCVERIFYREF and GCCREATEREF Macro
 * Transfered B2BUAConnection::OnIncomingCall to be called in B2BUAEndPoint::OnIncomingSIPMessage
 *
 * Revision 1.59  2007/05/24 14:58:36  joegenbaclor
 * Fixed bug in hold/unhold as reported by EarthLink
 *
 * Revision 1.58  2007/05/24 01:51:17  joegenbaclor
 * Added force-privacy=none support
 *
 * Revision 1.57  2007/05/18 10:47:33  joegenbaclor
 * Made sure xfer- is never concatenated twice for hairpinned calls
 *
 * Revision 1.56  2007/05/15 12:18:17  joegenbaclor
 * 1.  Added force-privacy option in route
 * 2.  Corrected some typos in last commit
 *
 * Revision 1.55  2007/05/11 13:00:32  joegenbaclor
 * Fixed bug in SendB2BReject where 407 is sent for 401 responses.  Thanks Cedric Lecourt
 *
 * Revision 1.54  2007/05/08 12:49:27  joegenbaclor
 * Corrected memory leaks in SIPMessage.
 *
 * Revision 1.53  2007/05/04 09:40:52  joegenbaclor
 * more on delayed deletion mechanism
 *
 * Revision 1.52  2007/05/01 04:18:42  rcolobong
 * 1. Bug Fix on calling OnSendB2BConnect after sending the request.
 *
 * Revision 1.51  2007/04/17 15:38:44  joegenbaclor
 * More work on local REFER.
 *
 * Revision 1.50  2007/04/10 06:13:07  joegenbaclor
 * Added new Encryption Mode
 *
 * Revision 1.49  2007/03/28 09:09:24  joegenbaclor
 * Logging enhancements
 *
 * Revision 1.48  2007/03/28 00:29:52  joegenbaclor
 * Modified loop detection code
 *
 * Revision 1.47  2007/03/27 08:55:09  joegenbaclor
 * added loop detection for incoming calls
 *
 * Revision 1.46  2007/03/02 06:22:46  joegenbaclor
 * Corrected some bug in GC.
 * Added CPU Throttle mechanism during heavy load to avoid denial of service
 * Made sure lists are Made unique in SIPHeade clone function
 *
 * Revision 1.45  2007/02/27 23:30:16  joegenbaclor
 * Made sure contact address in 200 OK use the same port used by the INVITE
 * Corrected minor typos from last commit
 *
 * Revision 1.44  2007/02/19 08:48:59  joegenbaclor
 * More work on privacy.
 * 1.  Change PAssertedIdentity base from SIPHeaderF to Contact to allow displaynames
 * 2.  Removed unused variables in ProfileUA
 * 3.  Allowed Privacy to be enabled via Makecall
 * 4.  Improved SIP session privacy support
 *
 * Revision 1.43  2007/02/16 11:09:24  joegenbaclor
 * More work on privacy
 *
 * Revision 1.42  2007/02/13 11:21:10  joegenbaclor
 * Made sure a call is destroyed and a 404 sent if the route can't be resolved
 *
 * Revision 1.41  2007/02/13 01:57:42  joegenbaclor
 * moved 200 ok retransmission right after SendRequest to avoid getting retransmitted
 *  without the inserted record route
 *
 * Revision 1.40  2007/02/08 18:05:13  joegenbaclor
 * Corrected another bug in Record routing
 *
 * Revision 1.39  2007/02/08 12:32:54  joegenbaclor
 * Reverted last TCP changes.  As per SIP Implementors, it is a bad idea to use the TCP outbound socket in the reverse direction
 *
 * Revision 1.38  2007/01/22 12:25:55  joegenbaclor
 * Added record route for B2B calls.
 *
 * Revision 1.37  2007/01/22 10:01:58  joegenbaclor
 * Fixed bug in handling Notifies
 *
 * Revision 1.36  2007/01/19 08:21:12  joegenbaclor
 * Modified Redirect behavior.  Second leg would now process redirect instead of the
 *  session relaying 3xx to UAC
 *
 * Revision 1.35  2007/01/10 08:07:16  rcolobong
 * 1. Add new method for OnSendB2BConnect Call Interface.
 * 2. OnSendB2BConnect is supported in B2BUserAgent, B2BUACall and B2BUAConnection
 *
 * Revision 1.34  2006/12/21 05:39:37  joegenbaclor
 * Fixed REPLACES to use connection to get the LEG 2 call-id
 *
 * Revision 1.33  2006/11/30 08:19:37  joegenbaclor
 * Relicensed opensipstack to MPL/GPL/LGPL triple license scheme
 *
 * Revision 1.32  2006/11/22 03:16:18  joegenbaclor
 * Removed Record Routes from B2B Outbound invites
 *
 * Revision 1.31  2006/11/22 00:05:52  joegenbaclor
 * Bug fixes for multi-listener support
 *
 * Revision 1.30  2006/11/21 06:55:19  joegenbaclor
 * Added ability to add/append additional listeners.
 *
 * Revision 1.29  2006/11/10 07:53:10  rcolobong
 * 1. Change Access modifier for method SetB2BUAConnection
 * 2. Support Fork Calls Messages by giving 1xx-6xx events to B2BUAForkCalls
 *
 * Revision 1.28  2006/11/08 06:46:20  joegenbaclor
 * Added session level NIC/Interface binding
 *
 * Revision 1.27  2006/11/07 07:50:32  joegenbaclor
 * Corrected bug in Session timer processing.
 *
 * Revision 1.26  2006/10/23 15:44:58  joegenbaclor
 * 1.  Numerous bug fixes for REFER
 * 2.  Added new SIP Headers for SIP timers and Attended call transfer
 * 3.  Added loop sanity check to SIP UDP Transport
 * 4.  Corrected parser bug for ProxyRequire header
 *
 * Revision 1.25  2006/10/11 04:49:47  rcolobong
 * 1. Add support for Notify During Refer
 * 2. Fixed bug where 3xx invoke CallHandler OnRejected
 *
 * Revision 1.24  2006/10/05 12:56:07  rcolobong
 * Change Redirection id which is "redir" to "rdir"
 *
 * Revision 1.23  2006/09/21 05:09:19  joegenbaclor
 * Corrected typo from last commit
 *
 * Revision 1.22  2006/09/21 04:43:18  joegenbaclor
 * Improved RTP logging
 *
 * Revision 1.21  2006/09/20 09:51:15  joegenbaclor
 * 1.  Disabled Media Aggregation in B2BMediaInterface
 * 2.  Added logging for RTP statitics for media proxy sessions
 *
 * Revision 1.20  2006/09/19 10:25:51  joegenbaclor
 * Added sanity check to B2BAnswerCall to check the CSeq method of the response for commiting
 *  to an AnswerCall()
 *
 * Revision 1.19  2006/09/18 09:35:16  joegenbaclor
 * Completed initial implementation of RTP media aggregation
 *
 * Revision 1.18  2006/09/08 04:39:17  joegenbaclor
 * Implemented INFO Session Keep-alive
 *
 * Revision 1.17  2006/09/07 09:49:58  joegenbaclor
 * Fixed bug on possible race conditions when multiple thread calls Session::Destroy
 *  at the same time
 *
 * Revision 1.16  2006/08/30 03:42:05  joegenbaclor
 * Corrected bug in garbage collector loop.
 *
 * Revision 1.15  2006/08/29 10:47:23  rcolobong
 * *** empty log message ***
 *
 * Revision 1.14  2006/08/28 07:18:35  joegenbaclor
 * Bulk Commit: Logging overhaul and Call Session callback changes to reflect SIPMessage eg:  OnAlerting(), OnProgress(), etc
 *
 * Revision 1.13  2006/08/26 14:31:44  joegenbaclor
 * Bulk Commit:  All SIP Sessions are now using new Garbage Collector and smart Pointers
 *
 * Revision 1.12  2006/07/19 02:19:21  joegenbaclor
 * 1.  Bug fixes concerning B2BUA destruction during during 3xx, 4xx, 5xx, 6xx local response
 * 2.  Fixed bug where Route header is never removed from outbound Invite resulting that may result to a loop when
 *      Inbound Invite has a Route header present
 * 3.  Added RemoveAllRoutes() in SIPMessage
 *
 * Revision 1.11  2006/07/17 11:36:45  joegenbaclor
 * 1.  More routing enhancements to B2BUA
 *
 * Revision 1.10  2006/07/13 06:53:05  joegenbaclor
 * 1.  Introduced Sanity checks for incoming SIP Message
 * 2.  Corrected bug in SIPURI where enum scheme is not getting recognized.
 * 3.  Added ENUM support to routing
 * 4.  Introduced global routing flag to indicate when and not to rewrite outbound To URI's
 *
 * Revision 1.9  2006/07/11 13:59:36  joegenbaclor
 * Completed Proxy functionalities for B2BUA
 *
 * Revision 1.8  2006/07/10 06:29:38  joegenbaclor
 * Completed crude Registration support for B2BUA
 *
 * Revision 1.7  2006/06/28 14:14:59  joegenbaclor
 * Bulk commit for Media Proxy functionalities for OpenB2BUA
 * Marks a minor version upgrade from 1.0.2 to 1.0.3
 *
 * Revision 1.6  2006/06/26 03:06:06  joegenbaclor
 * I have decided to include the latest development realease  of OPAL tagged Deimos Devel 1 (June 8 2006) as inegrated classes to opensipstack to
 * avoid future version conflicts due to the fast pace in OPAL development.   This move is also aimed to reduce the size of projects using OPAL
 * componets such as the soon to be relased OpenSIPPhone.
 *
 * Revision 1.5  2006/06/23 04:44:04  joegenbaclor
 * SendB2BRefer() added to B2BUACall
 *
 * Revision 1.4  2006/06/22 08:27:56  joegenbaclor
 * *** empty log message ***
 *
 * Revision 1.3  2006/06/21 14:29:11  joegenbaclor
 * This upload marks the first successful call by OpenB2BUA
 *
 * Revision 1.2  2006/06/20 09:52:55  joegenbaclor
 * Added Routing and SDP Handlers to B2BUA
 *
 * Revision 1.1  2006/06/15 09:47:28  joegenbaclor
 * More preliminary design work on B2BUA
 *
 *
 */

#include "B2BUA.h"
#include "B2BUACall.h"
#include "B2BUAConnection.h"
#include "B2BUAForkCalls.h"

#define new PNEW

using namespace B2BUA;

/// UAC constructor
B2BUACall::B2BUACall(
  B2BUAEndPoint & manager,
  const OString & sessionId,
  const ProfileUA & profile
) : CallSession( manager, sessionId, profile )
{
  m_LegIndex = 1;
#if HAS_OPAL
  m_HasClosedMediaStreams = FALSE;
  m_AudioRTPSessionThread = NULL;
  m_VideoRTPSessionThread = NULL;
  m_AudioAggregator = NULL;
  m_VideoAggregator = NULL;
#endif
  m_LogTag = "CALL: (outbound) ";

}

/// UAS constructor
B2BUACall::B2BUACall(
  B2BUAEndPoint & sessionManager,
  const SIPMessage & request,
  const OString & sessionId
) : CallSession( sessionManager, request, sessionId )
{
  m_LegIndex = 0;
#if HAS_OPAL
  m_HasClosedMediaStreams = FALSE;
  m_AudioRTPSessionThread = NULL;
  m_VideoRTPSessionThread = NULL;
  m_AudioAggregator = NULL;
  m_VideoAggregator = NULL;
#endif
  //SetLogContextId( sessionId + OString("-inbound") );

  m_LogTag = "CALL: (inbound) ";

}

B2BUACall::~B2BUACall()
{
  
}

void B2BUACall::SetB2BUAConnection( B2BUAConnection * conn )
{ 
  m_Connection = conn; 
}

void B2BUACall::CloseMediaStreams()
{
  PWaitAndSignal lock( m_MediaStreamsMutex );
  if( m_HasClosedMediaStreams )
    return;

  m_HasClosedMediaStreams = TRUE;

  RTP_UDP * audio = (RTP_UDP *)m_RTPSessionManager.GetSession( OpalMediaFormat::DefaultAudioSessionID );
  if( audio != NULL )
  {
    OStringStream stats;
    
    stats << "RTP: (Audio) ";
    stats << "addr=" << audio->GetLocalAddress() << ":" << audio->GetLocalDataPort() << "->";
    stats << audio->GetRemoteAddress() << ":" << audio->GetRemoteDataPort() << " ";
    stats << "enc=" << audio->IsEncryptionEnabled() << " ";
    stats << "rx=" << audio->GetPacketsReceived() << " ";
    stats << "tx=" << audio->GetPacketsSent() << " ";
    stats << "lost=" << audio->GetPacketsLost() << " ";
    stats << "outOfOrder=" << audio->GetPacketsOutOfOrder() << " ";
    stats << "late=" << audio->GetPacketsTooLate() << " ";

    stats << "rxTime=" << audio->GetAverageReceiveTime() << 
    "/" << audio->GetMaximumReceiveTime() << " ";

    stats << "txTime=" << audio->GetAverageSendTime() <<
    "/" << audio->GetMaximumSendTime() << " ";

    stats << "jitter=" << audio->GetAvgJitterTime() << 
    "/" << audio->GetMaxJitterTime();


    LOG( LogDetail(), stats.str() );

    audio->Close(TRUE);
    audio->Close(FALSE);
  }

  RTP_UDP * video = (RTP_UDP *)m_RTPSessionManager.GetSession( OpalMediaFormat::DefaultVideoSessionID );
  if( video != NULL )
  {
    OStringStream stats;
    stats << "RTP: (Video) ";
    
    stats << "addr=" << video->GetLocalAddress() << ":" << video->GetLocalDataPort() << "->";
    stats << video->GetRemoteAddress() << ":" << video->GetRemoteDataPort() << " ";
    stats << "enc=" << video->IsEncryptionEnabled() << " ";
    stats << "rx=" << video->GetPacketsReceived() << " ";
    stats << "tx=" << video->GetPacketsSent() << " ";
    stats << "lost=" << video->GetPacketsLost() << " ";
    stats << "outOfOrder=" << video->GetPacketsOutOfOrder() << " ";
    stats << "late=" << video->GetPacketsTooLate() << " ";

    stats << "rxTime=" << video->GetAverageReceiveTime() << 
    "/" << video->GetMaximumReceiveTime() << " ";

    stats << "txTime=" << video->GetAverageSendTime() <<
    "/" << video->GetMaximumSendTime() << " ";

    stats << "jitter=" << video->GetAvgJitterTime() << 
    "/" << video->GetMaxJitterTime();

    LOG( LogDetail(), stats.str() );

    video->Close(TRUE);
    video->Close(FALSE);
  }

}

void B2BUACall::OnDestroySession()
{
#if HAS_OPAL
  CloseMediaStreams();

  if( m_AudioRTPSessionThread != NULL )
  {
    PAssert( m_AudioRTPSessionThread->WaitForTermination( 10000 ), "Audio Session Thread Failed to terminate" );
    delete m_AudioRTPSessionThread;
  }
  
  if( m_VideoRTPSessionThread != NULL )
  {
    PAssert( m_VideoRTPSessionThread->WaitForTermination( 10000 ), "Video Session Thread Failed to terminate" );
    delete m_VideoRTPSessionThread;
  }

  if( m_AudioAggregator != NULL )
  {
    PAssert( m_AudioAggregator->WaitForTermination( 10000 ), "Audio Session Thread Failed to terminate" );
    m_AudioAggregator->GCCollect();
    m_AudioAggregator = NULL;
  }

  if( m_VideoAggregator != NULL )
  {
    PAssert( m_VideoAggregator->WaitForTermination( 10000 ), "Video Session Thread Failed to terminate" );
    m_VideoAggregator->GCCollect();
    m_VideoAggregator = NULL;
  }

#endif
}

void B2BUACall::OnSessionEvent(
  int event,
  const SIPMessage & eventMsg
)
{
  switch( event )
  {
    case IST_AnswerCallNow:                // Sends 200 OK back with SDP after receipt of INVITE
    if( SendB2BConnect( eventMsg ) )
      SetState( StateConnected );
      break;
    case IST_AnswerCallDeferred:           // Answers with 180 ringing
      //InternalAnswerCall( AnswerCallDeferred, eventMsg );
      SendB2BProgress( eventMsg );
      SetState( StateAlerting );
      break;
    case IST_AnswerCallDeferredWithMedia:  // Like AnswerCallDefered Only media is sent in 183 progress
      SendB2BProgress( eventMsg );
      SetState( StateProceeding );
      break;
    case IST_AnswerCallRedirect:           // Redirect the call to another address
#if 0
      SendB2BRedirect( eventMsg );
#endif
      SetState( StateRedirecting );
      break;
    case IST_AnswerCallRedirectToProxy:    // Redirect the call to a proxy
#if 0      
      SendB2BRedirect( eventMsg );
#endif
      SetState( StateRedirecting );
      break;
    case IST_AnswerCallDenied:              // Reject the call
      SendB2BReject( eventMsg );
      if( GetState() < StateConnected )
        SetState( StateDisconnected );
      break;
    default:
      CallSession::OnSessionEvent( event, eventMsg );
  }
}

BOOL B2BUACall::SendB2BConnect( 
  const SIPMessage & secondLegResponse 
)
{
  GCVERIFYREF( "B2BUACall::SendB2BConnect", FALSE );

  if( !secondLegResponse.Is2xx() )
    PAssertAlways( PLogicError );

  SIPMessage ok;
  if( !m_CurrentUASInvite.CreateResponse( ok, SIPMessage::Code200_Ok ) )
  {
    return SendReject( 
      m_CurrentUASInvite, 
      SIPMessage::Code500_InternalServerError
    );

    return FALSE;
  }


  OString sdp;
  if( secondLegResponse.HasSDP() )
  {
    if( !OnRequireSDPAnswer( m_CurrentUASInvite, ok ) )
    {
      PTRACE( 1, "B2BUACall::SendB2BConnect Rejecting Media" );

      SendReject( m_CurrentUASInvite, GetCallAnswerResponse() == DontDeny ?
                            SIPMessage::Code415_UnsupportedMedia :
          (SIPMessage::StatusCodes)GetCallAnswerResponse() );

          return FALSE;
    }
  }

/*
  PIPSocket::Address localAddress, targetAddress;
  WORD localPort = 0, targetPort = 0;
  SIPTransport::Resolve( ok, targetAddress, targetPort );
  GetSessionManager().GetTransportManager()->GetListenerAddress( 
    SIPTransport::UDP, targetAddress, localAddress, localPort );
*/


  /// set the contact address
  SIPURI curi;
  curi.SetHost( GetInterfaceAddress() );
  curi.SetPort( OString( GetInterfacePort() ) );
  ContactURI contactURI;
  contactURI.SetURI( curi );
  ok.AppendContact( contactURI );


  InitializeDialogAsUAS( m_CurrentUASInvite, ok );

  if( m_EnableSessionTimer && m_CurrentUASInvite.IsSupportedExtension( "timer" ) && !OnProcessSessionTimers( m_CurrentUASInvite, ok ) )
      return FALSE;  

  /// Let's add our record route just in case UAs would not honor our contact address
  /// proxies in the middle would be able to route back to us correctly

  RecordRoute recordRoute;
  RouteURI rr;
  rr.SetURI( curi );
  rr.SetLooseRouter( TRUE );
  recordRoute.AddURI( rr );
  ok.AppendRecordRoute( recordRoute );

  if( !GetB2BUAConnection().OnSendB2BConnect( *this, secondLegResponse ) )
    return FALSE;
  
  /// Here we should send the 200 ok to the retransmitter
  CallSessionManager::RetransmitThread * retran = (CallSessionManager::RetransmitThread *)m_CallSessionManager.GetRetransmitThread();
  retran->AddMessage( GetSessionId(), ok );
  /// Send it to the FSM once to terminate the IST transaction
  if( !SendRequest( ok ) )
    return FALSE;

  return TRUE;

}

BOOL B2BUACall::SendB2BProgress( 
  const SIPMessage & secondLegResponse 
)
{
  GCVERIFYREF( "B2BUACall::SendB2BProgress", FALSE );

  PAssert( secondLegResponse.Is1xx(), PLogicError );

  SIPMessage progress;
  if( !m_CurrentUASInvite.CreateResponse( 
    progress, 
    (SIPMessage::StatusCodes)secondLegResponse.GetStatusCode() ) )
    return FALSE;

  if( secondLegResponse.HasSDP() )
  {
    if( !OnRequireSDPAnswer( m_CurrentUASInvite, progress ) )
    {
      PTRACE( 1, "B2BUACall::SendB2BProgress Rejecting Media" );
      SendReject( m_CurrentUASInvite, GetCallAnswerResponse() == DontDeny ?
                             SIPMessage::Code415_UnsupportedMedia :
          (SIPMessage::StatusCodes)GetCallAnswerResponse() );
      return FALSE;
    }
  }

  InitializeDialogAsUAS( m_CurrentUASInvite, progress );

  m_CurrentUASInvite.SetToTag( progress.GetToTag() );

  return SendRequest( progress );
}

BOOL B2BUACall::SendB2BRedirect( 
  const SIPMessage & secondLegResponse 
)
{
  GCVERIFYREF( "B2BUACall::SendB2BRedirect", FALSE );

  PAssert( secondLegResponse.Is3xx(), PLogicError );

  To to;
  m_CurrentUASInvite.GetTo( to );
  SIPMessage redirect;
  m_CurrentUASInvite.CreateResponse( redirect, secondLegResponse.GetStatusCode() );
  

  if( secondLegResponse.GetContactSize() == 0 )
    return FALSE;

  OString userName = secondLegResponse.GetTopContactURI().GetUser();
  if( userName.IsEmpty() )
    return FALSE;

  OString redirectionId;
  m_Connection->OnRequireRedirectionId( *this, redirectionId );

  OStringStream redirectUser;
  redirectUser << "rdir-" << redirectionId << "-" << userName;

  PIPSocket::Address localAddress, targetAddress;
  WORD localPort = 0, targetPort = 0;

  SIPTransport::Resolve( redirect, targetAddress, targetPort );
  GetSessionManager().GetTransportManager()->GetListenerAddress( 
    SIPTransport::UDP, targetAddress, localAddress, localPort );

  /// set the contact address
  SIPURI curi;
  curi.SetHost( localAddress.AsSTLString() );
  curi.SetPort( OString( localPort ) );
  curi.SetUser( redirectUser.str() );
  ContactURI contactURI;
  contactURI.SetURI( curi );
  Contact contactHeader;
  contactHeader.AddURI( contactURI );
  ContactList contactList;
  contactList.Append( contactHeader );
  redirect.SetContactList( contactList );
  
  return SendRequest( redirect );
}

BOOL B2BUACall::SendB2BRefer(
  B2BUACall & referror,
  const SIPMessage & refer 
)
{
  ReferTo referTo;
  if( !refer.GetReferTo( referTo ) )
    return FALSE;

  OString userName = referTo.GetURI().GetUser();
  if( userName.IsEmpty() )
    return FALSE;

  SIPMessage referClone;
  if( !CloneRequestWithinDialog( refer, referClone ) )
    return FALSE;

  SIPURI referToURI = referTo.GetURI();
  OString referToUser = referToURI.GetUser();

  if( referToUser.Find( "xfer-" ) == P_MAX_INDEX )
  {
    OString xferId;
    m_Connection->OnRequireCallTransferId( referror, referTo, xferId );
    OStringStream xferUser;
    xferUser << "xfer-" << xferId << "-" << referToURI.GetUser();
    referToURI.SetUser( xferUser.str() );
  }else
  {
    referToURI.SetUser( referToUser );
  }

  /// see if it already resolves to us
  if( !GetTransportManager()->IsLocalAddressAndPort( referToURI ) )
  {
    referToURI.SetHost( m_LocalVia.GetAddress() );
    referToURI.SetPort( OString( m_LocalVia.GetPort() ) );
  }

  OString repl;
  if( referToURI.GetHeader( "replaces", repl ) )
  {
    Replaces replaces = ParserTools::UnescapeAsRFC2396( repl );

    /// find the connection for this callId.
    OString callId = replaces.GetCallId();

    SIPSession::GCRef gcRef = GetSessionManager().FindGCRefByCallId( callId );

    if( gcRef != NULL )
    {
      B2BUACall * call = dynamic_cast<B2BUACall *>( gcRef.GetObject() );
      if( call != NULL )
      {
        if( call->GetLegIndex() == 0 )
        {
          B2BUACall * replacesCall = call->GetB2BUAConnection().GetLeg2Call();
          if( call == NULL )
            return FALSE;

          SIPSession::GCRef callRef = GCCREATEREF( replacesCall, "B2BUACall::SendB2BRefer",
            "B2BUACall Leg2" );
          if( callRef == NULL )
            return FALSE;
          
          if( !callRef->ConstructReplaces( replaces ) )
            return FALSE;
          
          referToURI.SetHeader( "replaces", replaces.EscapeAsRFC2396() );
        } else if ( call->GetLegIndex() == 1 )
        {
          B2BUACall * replacesCall = call->GetB2BUAConnection().GetLeg1Call();
          if( call == NULL )
            return FALSE;

          SIPSession::GCRef callRef = GCCREATEREF( replacesCall, "B2BUACall::SendB2BRefer",
            "B2BUACall Leg1" );
          if( callRef == NULL )
            return FALSE;

          if( !callRef->ConstructReplaces( replaces ) )
            return FALSE;
          
          referToURI.SetHeader( "replaces", replaces.EscapeAsRFC2396() );
        }
      }
    }
  }

  referTo.SetURI( referToURI );
  referClone.SetReferTo( referTo );
 
  return SendRequest( referClone );
}

BOOL B2BUACall::SendB2BNotify(
  B2BUACall & call,
  const SIPMessage & notify
)
{
  SIPMessage notifyClone;
  if( !CloneRequestWithinDialog( notify, notifyClone ) )
    return FALSE;
  SIPMessage ok;

  /// respond with 200
  notify.CreateResponse( ok, SIPMessage::Code200_Ok );
  SendRequest( ok );

  return SendRequest( notifyClone );
}

BOOL B2BUACall::SendB2BReject(
  const SIPMessage & secondLegResponse 
)
{
  GCVERIFYREF( "B2BUACall::SendB2BReject", FALSE );

  if( secondLegResponse.GetStatusCode() == SIPMessage::Code407_ProxyAuthenticationRequired )
  {
    To to;
    m_CurrentUASInvite.GetTo( to );
    SIPMessage auth;
    m_CurrentUASInvite.CreateResponse( auth, SIPMessage::Code407_ProxyAuthenticationRequired );
    
    ProxyAuthenticate proxyAuth;
    if( secondLegResponse.GetProxyAuthenticate( proxyAuth ) )
      auth.SetProxyAuthenticate( proxyAuth );

    LOG_IF_DEBUG( LogDebugVeryHigh(), "*** REJECTED ***" << " INCOMING request " << endl << m_CurrentUASInvite << " with " << endl << auth );
    
    return SendRequest( auth );
  }else if( secondLegResponse.GetStatusCode() == SIPMessage::Code401_Unauthorized )
  {
    To to;
    m_CurrentUASInvite.GetTo( to );
    SIPMessage auth;
    m_CurrentUASInvite.CreateResponse( auth, SIPMessage::Code401_Unauthorized );
    
    WWWAuthenticate wwwAuth;
    if( secondLegResponse.GetWWWAuthenticate( wwwAuth ) )
      auth.SetWWWAuthenticate( wwwAuth );

    LOG_IF_DEBUG( LogDebugVeryHigh(), "*** REJECTED ***" << " INCOMING request " << endl << m_CurrentUASInvite << " with " << endl << auth );
    
    return SendRequest( auth );
  }

  return SendReject( 
    m_CurrentUASInvite, 
    secondLegResponse.GetStatusCode() 
  );
}

void B2BUACall::AnswerB2BCall( 
  B2BUACall & secondLeg,
  const SIPMessage & secondLegResponse
)
{
  GCVERIFYREF_VOID( "B2BUACall::AnswerB2BCall" );

  if( GetLegIndex() != 0 )
    PAssertAlways( PLogicError );

  PWaitAndSignal lock( m_AnswerCallMutex );

  if( secondLegResponse.GetCSeqMethod() *= "INVITE" )
  {
    if( secondLegResponse.Is1xx() )
    {
      if( secondLegResponse.HasSDP() )
      {
        EnqueueMessageEvent( 
          IST_AnswerCallDeferredWithMedia, 
          secondLegResponse );
      }else
      {
        EnqueueMessageEvent( 
          IST_AnswerCallDeferred, 
          secondLegResponse );
      }
    }else if( secondLegResponse.Is2xx() )
    {
      EnqueueMessageEvent( 
        IST_AnswerCallNow, 
        secondLegResponse );
    }else if( secondLegResponse.Is3xx() )
    {
      EnqueueMessageEvent( 
        IST_AnswerCallRedirect, 
        secondLegResponse );
    }else if( !secondLegResponse.IsRequest() ) // 4xx - 6xx
    {
      EnqueueMessageEvent( 
        IST_AnswerCallDenied, 
        secondLegResponse );
    }
  }else
  {
    LOG( LogError(), "Got  a none-INVITE response while call is proceeding or alerting!!!\n" << secondLegResponse );

  }

  return;
}

BOOL B2BUACall::OnIncomingSDPOffer(
  const SIPMessage & sdp
)
{
  GCVERIFYREF( "B2BUACall::OnIncomingSDPOffer", FALSE );
  return m_Connection->OnIncomingSDPOffer( *this, sdp );
}



/// your last chance to produce an answer to the offer.  
/// This will be called either before sending 183, 200 or ACK 
BOOL B2BUACall::OnRequireSDPAnswer(
  const SIPMessage & offer,
  SIPMessage & answer
)
{
  GCVERIFYREF( "B2BUACall::OnRequireSDPAnswer", FALSE );
  return m_Connection->OnRequireSDPAnswer( *this, offer, answer );
}

/// this will be called if an offer is not received in INVITE or when we are initiating the call
BOOL B2BUACall::OnRequireSDPOffer(
  SIPMessage & offer
)
{
  GCVERIFYREF( "B2BUACall::OnRequireSDPOffer", FALSE );
  return m_Connection->OnRequireSDPOffer( *this, offer );
}

void B2BUACall::OnSessionTimerExpire()
{
  GCVERIFYREF_VOID( "B2BUACall::OnSessionTimerExpire" );
  LOG( LogError(), "SIP Session Timer Expired!!!" );
  m_Connection->DestroyConnection(); 
}

/** IST CallBacks */
void B2BUACall::IST_OnReceivedInvite(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::IST_OnReceivedInvite" );
  PWaitAndSignal lock( m_AnswerCallMutex );

  OString toTag = msg.GetToTag();
  
  if( GetState() == StateConnected && !toTag.IsEmpty() )
  {
    IST_OnReceivedMoreInvite( msg );
    return;
  }
  
  /// check if the message has a to tagged 
  if( !toTag.IsEmpty() )
  {
    LOG( LogError(), "*** TO tag present.  Ignoring reINVITE without a dialog ***" );
    SetCallAnswerResponse( CallSession::DisconnectWithTransactionDoesNotExist );
    m_Connection->DestroyConnection( msg );
    return;
  }

  SIPMessage response;
  
  if( GetType() == SIPSession::Client )
  {
    /// oops, I think we just got the same INVITE that we just sent out
    msg.CreateResponse( response, SIPMessage::Code482_LoopDetected );
    LOG( LogError(), "*** LOOP DETECTED ***" );
    SendRequest( response );
    return;
  }
  
  m_CurrentUASInvite = msg;

  /// set the interface address to be used for future requests
  /// within the session;
  m_InterfaceAddress = msg.GetInterfaceAddress();
  m_InterfacePort = msg.GetInterfacePort();

  if( !OnIncomingConnection( msg ) )
  {
    LOG_IF_DEBUG( LogError(), "OnIncomingSession FAILED.  Rejecting incoming INVITE." );

    SetCallAnswerResponse( GetCallAnswerResponse() == DontDeny ?
      CallSession::DisconnectWithNotFound : GetCallAnswerResponse());
    
    m_Connection->DestroyConnection( msg );
    return;
  }

  if( !OnSetupIncomingCall( msg ) )
  {
    LOG_IF_DEBUG( LogError(), "OnIncomingSession FAILED.  Rejecting incoming INVITE." );

    SetCallAnswerResponse( GetCallAnswerResponse() == DontDeny ?
      CallSession::DisconnectWithNotFound : GetCallAnswerResponse());
    
    m_Connection->DestroyConnection( msg );
    return;
  }
  
  /// We CAN'T answer the call at this point
  /// InternalAnswerCall( m_CallSessionManager.OnAnswerCall( msg, *this ), msg );

  m_Connection->EnqueueSessionEvent( new SIPSessionEvent( *m_Connection, B2BUAConnection::SetupOutbound, msg ) );
}

void B2BUACall::OnSetupOutbound(
  const SIPMessage & inboundInvite,
  B2BUAConnection & connection
)
{
  GCVERIFYREF_VOID( "B2BUACall::OnSetupOutbound" );
  
  B2BRoutingInterface::B2BRouteList routes;
  if( connection.IsUpstreamProxy() 
    && connection.GetUpstreamProxyRoutes().IsEmpty() )
  {
    routes = connection.GetUpstreamProxyRoutes();
  } else
  {
    routes = connection.GetRoutes();
  }

  if( routes.GetSize() == 0 )
  {
    SIPMessage noRoute;
    inboundInvite.CreateResponse( noRoute, SIPMessage::Code404_NotFound );
    connection.SendRequest( noRoute );

    connection.EnqueueSessionEvent( 
      new SIPSessionEvent( connection, B2BUAConnection::ErrorNoRoute ) );
    return;
  }

  PIPSocket::Address targetAddress;
  WORD targetPort;
  SIPURI uri( routes[0] );
  
  if( !SIPTransport::Resolve( uri, targetAddress, targetPort ) )
  {
    LOG_IF_DEBUG( LogWarning(), uri << " does not resolve to an IP Address" );
    SIPMessage noRoute;
    inboundInvite.CreateResponse( noRoute, SIPMessage::Code404_NotFound );
    connection.SendRequest( noRoute );

     connection.EnqueueSessionEvent( 
      new SIPSessionEvent( connection, B2BUAConnection::ErrorRouteResolution ) );
    return;
  }

  //m_SessionProfile.SetRemoteURI( uri.AsString() );
  SIPMessage invite = inboundInvite;

  
  /// IncomingConnection Allowed this call to process
  /// check for presence of SDP

  if( SIPTransport::IsPrivateNetwork( targetAddress ) )
  {
    connection.SetMediaProxyIfPrivate( FALSE );
  }
  
  BOOL hasSDP = invite.HasSDP();
  OString sdpOffer;

  if( hasSDP )
  {
    OString sdp;
    invite.GetBody( sdp );
    sdpOffer = OString( sdp );

    
    if( !m_Connection->GetLeg1Call()->OnIncomingSDPOffer( invite ) )
    {
      LOG_IF_DEBUG( LogError(), "OnIncomingSDPOffer FAILED.  Rejecting incoming INVITE." );


      SetCallAnswerResponse( GetCallAnswerResponse() == DontDeny ?
        CallSession::DisconnectWithUnsupportedMediaType : GetCallAnswerResponse());
    
      m_Connection->DestroyConnection( invite );

      return;
    }
  }
  
  invite.RemoveAllRoutes();

  invite.ClearInternalHeaders();

  BOOL willUseProxy = FALSE;

  /// create the request line
  RequestLine requestLine;
  requestLine.SetMethod( "INVITE" );
  willUseProxy = !m_SessionProfile.GetProxyProfile().GetServerAddress().IsEmpty();
  requestLine.SetRequestURI( uri );
  invite.SetStartLine( requestLine );

  /// set the to uri;
  if( connection.WillRewriteToURI() )
  {
    To to;
    invite.GetTo( to );
    to.SetURI( uri );
    invite.SetTo( to );
  }
  
 
  if( willUseProxy )
  {
    /// put the proxy as the static route
    SIPURI sendAddress( m_SessionProfile.GetProxyProfile().GetServerAddress() );
    LOG( LogInfo(), "B2BUACall will use Proxy: " << sendAddress );
    invite.SetSendAddress( sendAddress );
  }

  OString transport;
  if( !uri.GetParameter( "transport", transport ) )
    transport = "udp";

  
  Via via;
  GetSessionManager().ConstructVia( targetAddress, via, SIPTransport::UDP, connection.GetInterfacePort() );
  via.SetBranch( ParserTools::GenBranchParameter() );
  via.AddParameter( "rport", "" );
  invite.RemoveAllVias();
  invite.AppendVia( via );

  /// Set the inteface address to be used for future requests
  /// based it from the via host
  m_InterfaceAddress = via.GetAddress();
  m_InterfacePort = via.GetPort();

  /// Set the call Id
  CallId callId;
  callId.SetHeaderBody( GetSessionId() );
  invite.SetCallId( callId );

  /// Set the contact
  SIPURI contactURI;
  contactURI.SetHost( via.GetURI().GetHost() );
  contactURI.SetPort( via.GetURI().GetPort() );
  ContactURI contact( contactURI, m_SessionProfile.GetDisplayName() );
  invite.RemoveAllContact();
  invite.AppendContact( contact );

  
  OnRequireSDPOffer( invite );
  
  SetRequest( invite );
 // m_CurrentUACInvite = invite;

  /// check if the route is to use an XOR Hash
  OString paramEnc = uri.GetParameter( "encrypted" );
  if( paramEnc.ToLower() == "true" )
  {
    invite.SetEncryption( TRUE );
    EnableEncryption();
  }

  /// check if we need to force privacy
  OString paramPrivacy = uri.GetParameter( "force-privacy" ).ToLower();
  if( paramPrivacy == "id" || paramPrivacy == "true"  )
  {
    LOG( LogInfo(), "Enforcing Privacy for outbound" );

    PAssertedIdentity identity;
    ContactURI puri( invite.GetFromURI() );
    identity.AddURI( puri );
    invite.AppendPAssertedIdentity( identity );
    
    From from( "From: sip:anonymous@anonymous.invalid" );
    if( !invite.GetFromTag().IsEmpty() )
      from.AddParameter( "tag", invite.GetFromTag() );
    invite.SetFrom( from );

    Privacy privacy( "Privacy: id" );
    invite.SetPrivacy( privacy );
  }else if( paramPrivacy == "none" )
  {
    LOG( LogInfo(), "Enforcing Privacy for outbound" );

    PAssertedIdentity identity;
    ContactURI puri( invite.GetFromURI() );
    identity.AddURI( puri );

    invite.AppendPAssertedIdentity( identity );
    
    Privacy privacy( "Privacy: none" );
    invite.SetPrivacy( privacy );
  }

  invite.RemoveAllRecordRoutes();

  connection.OnOutgoingCall( *this, invite );
  m_CurrentUACInvite = invite;

  SendRequest( invite );
}

void B2BUACall::OnSessionKeepAliveTimeout()
{
  m_Connection->OnSessionKeepAliveTimeout( *this );
}


void B2BUACall::IST_OnReceivedMoreInvite(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::IST_OnReceivedMoreInvite" );
#if 0
  RTP_SessionManager & rtpManager = GetRTPSessionManager();
  RTP_UDP * rtp = (RTP_UDP *)rtpManager.GetSession( 
  OpalMediaFormat::DefaultAudioSessionID );
  if( rtp != NULL )
    CallSession::IST_OnReceivedMoreInvite( msg );
#endif
  m_CurrentUASInvite = msg;

  BOOL hasSDP = msg.HasSDP();
  OString sdpOffer;

  if( hasSDP )
    OnIncomingSDPOffer( msg );

  m_Connection->IST_OnReceivedMoreInvite( *this, msg );
}

void B2BUACall::IST_OnReceivedAck(
  const SIPMessage & msg
)
{
  CallSession::IST_OnReceivedAck( msg );
}

void B2BUACall::IST_OnReceivedUnknown(
  const SIPMessage & msg
)
{
  CallSession::IST_OnReceivedUnknown( msg );
}

/** NIST Callbacks */
void B2BUACall::NIST_OnReceivedRegister(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NIST_OnReceivedRegister" );
  CallSession::NIST_OnReceivedRegister( msg );
  m_Connection->NIST_OnReceivedRegister( *this, msg );
}

void B2BUACall::NIST_OnReceivedRefer(
  const SIPMessage & msg
)
{
  /// don't call CallSession::NIST_OnReceivedRefer() here
  /// we want a pure pass through behavior for REFER
  GCVERIFYREF_VOID( "B2BUACall::NIST_OnReceivedRefer" );
  m_Connection->NIST_OnReceivedRefer( *this, msg );
}

void B2BUACall::NIST_OnReceivedBye(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NIST_OnReceivedBye" );
  CallSession::NIST_OnReceivedBye( msg );
  m_Connection->NIST_OnReceivedBye( *this, msg );
}

void B2BUACall::NIST_OnReceivedOptions(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NIST_OnReceivedOptions" );
  CallSession::NIST_OnReceivedOptions( msg );
  m_Connection->NIST_OnReceivedOptions( *this, msg );
}

void B2BUACall::NIST_OnReceivedInfo(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NIST_OnReceivedInfo" );
  m_Connection->NIST_OnReceivedInfo( *this, msg );
}

void B2BUACall::NIST_OnReceivedMessage(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NIST_OnReceivedMessage" );
  m_Connection->NIST_OnReceivedMessage( *this, msg );
}

void B2BUACall::NIST_OnReceivedCancel(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NIST_OnReceivedCancel" );
  CallSession::NIST_OnReceivedCancel( msg );
  m_Connection->NIST_OnReceivedCancel( *this, msg );
}

void B2BUACall::NIST_OnReceivedNotify(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NIST_OnReceivedNotify" );
  CallSession::NIST_OnReceivedNotify( msg );
  m_Connection->NIST_OnReceivedNotify( *this, msg );
}

void B2BUACall::NIST_OnReceivedSubscribe(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NIST_OnReceivedSubscribe" );
  CallSession::NIST_OnReceivedSubscribe( msg );
  m_Connection->NIST_OnReceivedSubscribe( *this, msg );
}

void B2BUACall::NIST_OnReceivedUnknown(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NIST_OnReceivedUnknown" );
  CallSession::NIST_OnReceivedUnknown( msg );
  m_Connection->NIST_OnReceivedUnknown( *this, msg );
}

/** NICT CALLBACKS */
void B2BUACall::NICT_OnReceived1xx(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NICT_OnReceived1xx" );
  CallSession::NICT_OnReceived1xx( msg );
  m_Connection->NICT_OnReceived1xx( *this, msg );
}

void B2BUACall::NICT_OnReceived2xx(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NICT_OnReceived2xx" );
  CallSession::NICT_OnReceived2xx( msg );
  m_Connection->NICT_OnReceived2xx( *this, msg );
}

void B2BUACall::NICT_OnReceived3xx(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NICT_OnReceived3xx" );
  CallSession::NICT_OnReceived3xx( msg );
  m_Connection->NICT_OnReceived3xx( *this, msg );
}

void B2BUACall::NICT_OnReceived4xx(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NICT_OnReceived4xx" );
  CallSession::NICT_OnReceived4xx( msg );
  m_Connection->NICT_OnReceived4xx( *this, msg );
}

void B2BUACall::NICT_OnReceived5xx(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NICT_OnReceived5xx" );
  CallSession::NICT_OnReceived5xx( msg );
  m_Connection->NICT_OnReceived5xx( *this, msg );
}

void B2BUACall::NICT_OnReceived6xx(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NICT_OnReceived6xx" );
  CallSession::NICT_OnReceived6xx( msg );
  m_Connection->NICT_OnReceived6xx( *this, msg );
}

void B2BUACall::NICT_OnReceivedUnknown(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::NICT_OnReceivedUnknown" );
  CallSession::NICT_OnReceivedUnknown( msg );
  m_Connection->NICT_OnReceivedUnknown( *this, msg );
}

/** ICT CALLBACKS */
void B2BUACall::ICT_OnReceived1xx(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::ICT_OnReceived1xx" );
  CallSession::ICT_OnReceived1xx( msg );

  if( m_Connection->IsForkCalls() )
  {
    B2BUAForkCalls * forkCalls = m_Connection->GetForkCalls();
    if( forkCalls )
    {
      if( !forkCalls->OnReceived1xxResponse( *this, msg ) )
        return;
    }
  }

  m_Connection->ICT_OnReceived1xx( *this, msg );
}

void B2BUACall::ICT_OnReceived2xx(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::ICT_OnReceived2xx" );

  DWORD cseqNo = msg.GetCSeqNumber();
  if( cseqNo <= (DWORD)m_LastICTResponseCSeq )
  {
    LOG( LogDebug(), "Dropping 200 OK.  CSeq " << cseqNo << " is <= current " << m_LastICTResponseCSeq );
    if( cseqNo == (DWORD)m_LastICTResponseCSeq ) // this is a retransmission
      SendAck( msg );
    return;  
  }
  
  CallSession::ICT_OnReceived2xx( msg );

  if( m_Connection->IsForkCalls() )
  {
    B2BUAForkCalls * forkCalls = m_Connection->GetForkCalls();
    if( forkCalls )
    {
      if( !forkCalls->OnReceived2xxResponse( *this, msg ) )
        return;
    }
  }

  m_Connection->ICT_OnReceived2xx( *this, msg );
}

void B2BUACall::ICT_OnReceived3xx(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::ICT_OnReceived3xx" );

  if( m_Connection->IsForkCalls() )
  {
    B2BUAForkCalls * forkCalls = m_Connection->GetForkCalls();
    if( forkCalls )
    {
      if( forkCalls->OnReceivedErrorResponses( *this, msg ) )
        return;
    }
  }

  m_Connection->ICT_OnReceived3xx( *this, msg );
  // the default behaviour would try to process the redirect internally
  CallSession::ICT_OnReceived3xx( msg );


}

void B2BUACall::ICT_OnReceived4xx(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::ICT_OnReceived4xx" );
  //CallSession::ICT_OnReceived4xx( msg );

  if( m_Connection->IsForkCalls() )
  {
    B2BUAForkCalls * forkCalls = m_Connection->GetForkCalls();
    if( forkCalls )
    {
      forkCalls->OnReceivedErrorResponses( *this, msg );
      return;
    }
  }

  m_Connection->ICT_OnReceived4xx( *this, msg );
}

void B2BUACall::ICT_OnReceived5xx(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::ICT_OnReceived5xx" );
  CallSession::ICT_OnReceived5xx( msg );

  if( m_Connection->IsForkCalls() )
  {
    B2BUAForkCalls * forkCalls = m_Connection->GetForkCalls();
    if( forkCalls )
    {
      forkCalls->OnReceivedErrorResponses( *this, msg );
      return;
    }
  }

  m_Connection->ICT_OnReceived5xx( *this, msg );
}

void B2BUACall::ICT_OnReceived6xx(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::ICT_OnReceived6xx" );
  CallSession::ICT_OnReceived6xx( msg );

  if( m_Connection->IsForkCalls() )
  {
    B2BUAForkCalls * forkCalls = m_Connection->GetForkCalls();
    if( forkCalls )
    {
      forkCalls->OnReceivedErrorResponses( *this, msg );
      return;
    }
  }

  m_Connection->ICT_OnReceived6xx( *this, msg );
}

void B2BUACall::ICT_OnReceivedUnknown(
  const SIPMessage & msg
)
{
  GCVERIFYREF_VOID( "B2BUACall::ICT_OnReceivedUnknown" );
  CallSession::ICT_OnReceivedUnknown( msg );
  m_Connection->ICT_OnReceivedUnknown( *this, msg );
}




