/*
 *
 * B2BUAConnection.cxx
 * 
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: B2BUAConnection.cxx,v $
 * Revision 1.87  2007/07/26 11:52:56  joegenbaclor
 * Finalized AuthenticationPending
 *
 * Revision 1.86  2007/07/26 08:15:53  joegenbaclor
 * Uploading support for AuthenticationPending state in B2BUAConnection.  This fix will eliminatethe need  to generate a diffrent call-id for resending credentials for 401/407
 *
 * Revision 1.85  2007/07/15 13:33:12  joegenbaclor
 * deprecated old connection count variable
 *
 * Revision 1.84  2007/07/14 07:50:29  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.83  2007/07/10 03:23:05  joegenbaclor
 * Added experimental code to use STL tokenizer instead of internal string tokenizer
 *
 * Revision 1.82  2007/07/06 13:38:39  rcolobong
 * Add Upstream Proxy Route feature
 *
 * Revision 1.81  2007/06/16 12:30:40  joegenbaclor
 * Changed CacheManager to return a smart pointer instead of a direct pointer to cache object to avoid segmentation faults when an object expires in cache and another thread has pending ownership of the pointer
 *
 * Renamed CacheManager::GetAt() to CacheManager::CreatePointer() to give a hint to users that the pointer needs to be deleted after being used
 *
 * Revision 1.80  2007/06/15 09:12:14  joegenbaclor
 * Added IVR Transfer reject event
 *
 * Revision 1.79  2007/06/13 09:22:05  joegenbaclor
 * Implemented basic IVR handler in OpenSBC
 *
 * Revision 1.78  2007/06/13 04:16:49  joegenbaclor
 * Fixed bug in IVR call transfer
 *
 * Revision 1.77  2007/06/12 09:22:41  joegenbaclor
 * Added IVR Call Transfer
 *
 * Revision 1.76  2007/06/08 15:10:32  joegenbaclor
 * Added IVR DTMF Grammar
 *
 * Revision 1.75  2007/06/08 09:26:30  joegenbaclor
 * Added B2BIVRInterface handlers
 *
 * Revision 1.74  2007/06/08 02:33:46  joegenbaclor
 * Commiting more Media Server Related Changes
 *
 * Revision 1.73  2007/06/06 09:54:15  joegenbaclor
 * More B2BUAConnection IVR methods
 *
 * Revision 1.72  2007/06/05 16:27:56  joegenbaclor
 * Added IVR audio playback functions to B2BUAConnection
 *
 * Revision 1.71  2007/05/27 06:15:50  joegenbaclor
 * Implemented OnReceivedSDPAnswer()
 *
 * Revision 1.70  2007/05/25 08:38:09  joegenbaclor
 * Introduced GCVERIFYREF and GCCREATEREF Macro
 * Transfered B2BUAConnection::OnIncomingCall to be called in B2BUAEndPoint::OnIncomingSIPMessage
 *
 * Revision 1.69  2007/05/24 14:58:36  joegenbaclor
 * Fixed bug in hold/unhold as reported by EarthLink
 *
 * Revision 1.68  2007/05/16 08:11:15  rcolobong
 * 1. Bug fix on Alerting and Seize Timeout
 * 2. Bug fix on AttachCall method where existing B2BUACall is overriden without releasing its reference
 *
 * Revision 1.67  2007/05/08 05:03:25  joegenbaclor
 * Added 491 Request Pending Support
 *
 * Revision 1.66  2007/05/07 16:15:01  joegenbaclor
 * Fixed memory leaks in Garbage Collection
 *
 * Revision 1.65  2007/05/06 15:52:14  joegenbaclor
 * More bug fixes in Delayed Deletion
 *
 * Revision 1.64  2007/05/04 09:40:52  joegenbaclor
 * more on delayed deletion mechanism
 *
 * Revision 1.63  2007/05/04 08:44:42  joegenbaclor
 * Introduced delayed deletion mechanism
 *
 * Revision 1.62  2007/05/01 04:19:37  rcolobong
 * 1. Add support for enabling and disabling Local Refer
 * 2. Invoke OnCallRejected during failover
 * 3. Add method GenerateCallIdExtensions
 *
 * Revision 1.61  2007/04/20 06:36:50  joegenbaclor
 * Completed initial work on local REFER
 *
 * Revision 1.60  2007/04/19 04:21:09  joegenbaclor
 * Big fixes OnHandleStateTransferring
 *
 * Revision 1.59  2007/04/18 11:10:21  joegenbaclor
 * Corrected type in last commit
 *
 * Revision 1.58  2007/04/18 10:56:29  joegenbaclor
 * Ignored sip messages received froma referred leg when call is already connected.
 *
 * Revision 1.57  2007/04/18 05:35:41  joegenbaclor
 * Added method to force exit of media threads for REFERred calls
 *
 * Revision 1.56  2007/04/17 15:38:44  joegenbaclor
 * More work on local REFER.
 *
 * Revision 1.55  2007/04/17 08:02:15  joegenbaclor
 * more work on local refer
 *
 * Revision 1.54  2007/04/16 14:24:41  joegenbaclor
 * More work on B2B local REFER support
 *
 * Revision 1.53  2007/04/09 08:06:20  joegenbaclor
 * Fixed bug in ACK transaction matching
 *
 * Revision 1.52  2007/03/02 06:22:46  joegenbaclor
 * Corrected some bug in GC.
 * Added CPU Throttle mechanism during heavy load to avoid denial of service
 * Made sure lists are Made unique in SIPHeade clone function
 *
 * Revision 1.51  2007/02/16 11:09:25  joegenbaclor
 * More work on privacy
 *
 * Revision 1.50  2007/02/13 11:21:10  joegenbaclor
 * Made sure a call is destroyed and a 404 sent if the route can't be resolved
 *
 * Revision 1.49  2007/02/09 03:47:35  joegenbaclor
 * 1. Added MaxForwards handling for B2B calls
 * 2. #defined SIP_DEFAULT_MAX_FORWARDS constant
 *
 * Revision 1.48  2007/02/08 18:05:13  joegenbaclor
 * Corrected another bug in Record routing
 *
 * Revision 1.47  2007/01/22 10:01:58  joegenbaclor
 * Fixed bug in handling Notifies
 *
 * Revision 1.46  2007/01/19 08:21:12  joegenbaclor
 * Modified Redirect behavior.  Second leg would now process redirect instead of the
 *  session relaying 3xx to UAC
 *
 * Revision 1.45  2007/01/10 08:07:16  rcolobong
 * 1. Add new method for OnSendB2BConnect Call Interface.
 * 2. OnSendB2BConnect is supported in B2BUserAgent, B2BUACall and B2BUAConnection
 *
 * Revision 1.44  2006/12/19 10:14:26  rcolobong
 * 1. Add method to insert Warning header in 200 Ok response to inbound
 * 2. Alerting timeout will now do failover
 *
 * Revision 1.43  2006/12/17 13:30:58  joegenbaclor
 * Added call attempt index in call-id string to avoid call-id collision during failovers
 *
 * Revision 1.42  2006/11/30 08:19:37  joegenbaclor
 * Relicensed opensipstack to MPL/GPL/LGPL triple license scheme
 *
 * Revision 1.41  2006/11/22 00:05:52  joegenbaclor
 * Bug fixes for multi-listener support
 *
 * Revision 1.40  2006/11/16 01:29:33  rcolobong
 * Notify CallHandler during Authentication Required
 *
 * Revision 1.39  2006/11/10 07:55:35  rcolobong
 * 1. Support Fork Invites using B2BUAForkCalls
 * 2. Handle Cancel events during Forking
 *
 * Revision 1.38  2006/10/30 08:35:55  joegenbaclor
 * Removed more references to presence namespace
 *
 * Revision 1.37  2006/10/25 09:07:46  rcolobong
 * Fix bug where leg1 and leg2 should be set to NULL after OnDestroySession
 *
 * Revision 1.36  2006/10/23 15:44:58  joegenbaclor
 * 1.  Numerous bug fixes for REFER
 * 2.  Added new SIP Headers for SIP timers and Attended call transfer
 * 3.  Added loop sanity check to SIP UDP Transport
 * 4.  Corrected parser bug for ProxyRequire header
 *
 * Revision 1.35  2006/10/11 04:49:47  rcolobong
 * 1. Add support for Notify During Refer
 * 2. Fixed bug where 3xx invoke CallHandler OnRejected
 *
 * Revision 1.34  2006/10/05 13:03:54  rcolobong
 * 1. Fix bug regarding OnRequireCallTransferId calling B2BUA OnRequireRedirectionId Method
 * 2. Add flag in OnSetupOutbound to check if Connection State is already disconnected
 *
 * Revision 1.33  2006/09/25 04:43:30  rcolobong
 * 1. Fix bug in 415 Unsupported Media where CallSession::m_ICTState is not properly initialized during construction.
 * 2. Remove the temporary fix for the 415 Unsupported Media where we use Semaphore to wait for second leg SDP in OnRequireSDPAnswer
 *
 * Revision 1.32  2006/09/22 08:52:45  joegenbaclor
 * Temporary fix for OnRequireSDPAnswer to use a semaphore to wait for second leg SDP
 *
 * Revision 1.31  2006/09/20 11:05:26  joegenbaclor
 * Flagged sending of Session Keep Alive.  Defaults to FALSE
 *
 * Revision 1.30  2006/09/19 10:47:12  rcolobong
 * Disable temporarily session keep alive
 *
 * Revision 1.29  2006/09/18 09:35:16  joegenbaclor
 * Completed initial implementation of RTP media aggregation
 *
 * Revision 1.28  2006/09/18 03:28:55  joegenbaclor
 * 1. More work on media aggregation
 * 2. Check sock->GetPort() for NULL in SIPTransportManager::GetListenerAddress() (thanks to
 *  Vamsi Pottangi for reporting the bug)
 * 3.  Added reference size check to GCObject::ReleaseRef()
 *
 * Revision 1.27  2006/09/15 15:31:39  joegenbaclor
 * Changed Log levels to decrease log size.
 * Changed default session keep alaive interval from 10 seconds to two minutes
 *
 * Revision 1.26  2006/09/15 14:45:22  joegenbaclor
 * Flagged SetSessionState of B2BConnection so it wont propagate events when connection
 *  is already destroyed
 *
 * Revision 1.25  2006/09/15 12:29:45  joegenbaclor
 * Flagged OnDestroySession of B2BConnection so it wont be called twice from different
 *  threads.
 *
 * Revision 1.24  2006/09/08 04:39:17  joegenbaclor
 * Implemented INFO Session Keep-alive
 *
 * Revision 1.23  2006/09/02 17:58:20  joegenbaclor
 * Corrected compile errors in GCC regarding euqation of NULL to GCRef.
 *
 * Revision 1.22  2006/09/01 13:15:58  rcolobong
 * Add Logging during Construction and Destruction of B2BUAConnection
 *
 * Revision 1.21  2006/08/30 08:45:41  rcolobong
 * 1. Update handling for connection and alerting timeout
 * 2. Fix bug where leg1 is destroyed before calling DestroyConnection
 *
 * Revision 1.20  2006/08/30 03:42:05  joegenbaclor
 * Corrected bug in garbage collector loop.
 *
 * Revision 1.19  2006/08/29 10:47:23  rcolobong
 * *** empty log message ***
 *
 * Revision 1.18  2006/08/28 07:18:35  joegenbaclor
 * Bulk Commit: Logging overhaul and Call Session callback changes to reflect SIPMessage eg:  OnAlerting(), OnProgress(), etc
 *
 * Revision 1.17  2006/08/26 14:31:44  joegenbaclor
 * Bulk Commit:  All SIP Sessions are now using new Garbage Collector and smart Pointers
 *
 * Revision 1.16  2006/08/16 04:14:33  rcolobong
 * Invoke StopConnectionTimeout after receiving 2xx response
 *
 * Revision 1.15  2006/08/04 05:31:29  rcolobong
 * 1.  Add Alerting and Connecting Timer
 * 2.  Support failover
 * 3.  Add Alerting, CallAbandoned, Rejected, Disconnected, Connected
 *
 * Revision 1.14  2006/07/19 02:19:21  joegenbaclor
 * 1.  Bug fixes concerning B2BUA destruction during during 3xx, 4xx, 5xx, 6xx local response
 * 2.  Fixed bug where Route header is never removed from outbound Invite resulting that may result to a loop when
 *      Inbound Invite has a Route header present
 * 3.  Added RemoveAllRoutes() in SIPMessage
 *
 * Revision 1.13  2006/07/13 06:53:05  joegenbaclor
 * 1.  Introduced Sanity checks for incoming SIP Message
 * 2.  Corrected bug in SIPURI where enum scheme is not getting recognized.
 * 3.  Added ENUM support to routing
 * 4.  Introduced global routing flag to indicate when and not to rewrite outbound To URI's
 *
 * Revision 1.12  2006/07/06 05:38:11  joegenbaclor
 * Final last minute changes to CBCOM hash support
 *
 * Revision 1.11  2006/06/29 05:09:13  joegenbaclor
 * Changed OnOutgoingCall invite parameter from const to none const to give a
 * chance for applications to modify the outbound invite before being sent to the transport
 *
 * Revision 1.10  2006/06/28 14:14:59  joegenbaclor
 * Bulk commit for Media Proxy functionalities for OpenB2BUA
 * Marks a minor version upgrade from 1.0.2 to 1.0.3
 *
 * Revision 1.9  2006/06/23 04:44:04  joegenbaclor
 * SendB2BRefer() added to B2BUACall
 *
 * Revision 1.8  2006/06/22 08:27:57  joegenbaclor
 * *** empty log message ***
 *
 * Revision 1.7  2006/06/21 14:29:11  joegenbaclor
 * This upload marks the first successful call by OpenB2BUA
 *
 * Revision 1.6  2006/06/20 09:52:55  joegenbaclor
 * Added Routing and SDP Handlers to B2BUA
 *
 * Revision 1.5  2006/06/15 09:47:28  joegenbaclor
 * More preliminary design work on B2BUA
 *
 * Revision 1.4  2006/06/14 08:43:38  joegenbaclor
 * Initial upload of OpenB2BUA applcation and related classes
 *
 * Revision 1.3  2006/04/11 00:18:54  joegenbaclor
 * More work on B2BUA class structures
 *
 * Revision 1.2  2006/04/10 01:09:03  joegenbaclor
 * Implemented Create methods fo B2BUA sessions
 *
 * Revision 1.1  2006/04/08 06:27:13  joegenbaclor
 * Initial upload of B2BUA classes
 *
 *
 */

#include "B2BUAConnection.h"
#include "B2BUAForkCalls.h"
#include "ResourceCounter.h"

#define new PNEW

using namespace B2BUA;

DWORD B2BUAConnection::m_SeizeTimeout = 0;
DWORD B2BUAConnection::m_AlertingTimeout = 0;

B2BUAConnection::B2BUAConnection(
  B2BUAEndPoint & ep,
  const OString & sessionId
) : SIPSession( ep, sessionId )
{
  m_LogTag = "CONNECTION";
  m_RewriteToURI = FALSE;

  m_Leg1Call = NULL;
  m_Leg2Call = NULL;
  m_ForkCalls = new B2BUAForkCalls( *this );

  m_SessionState = Idle;
  m_CallSessions.DisallowDeleteObjects();

  m_MediaProxyIfPrivate = FALSE;  /// Always Proxy Media by default
  m_EnableForkCalls = FALSE;
  m_EnableUpstreamProxyRoute = FALSE;

  LOG_CONTEXT( LogInfo(), sessionId, "B2BUAConnection Created 0x" << std::hex << this );
  OStringStream rc;
  ResourceCounter::Print( rc );
  LOG_CONTEXT( LogInfo(), sessionId, "*** COUNTERS *** " << rc.str() );

#if HAS_OPAL
  m_HasStartedMediaStream = FALSE;
#endif

  m_ConnectionDestroyed = FALSE;
  m_SendSessionKeepAlive = FALSE;
  m_HasReceivedSDP = FALSE;
  m_EnableLocalRefer = FALSE;
  m_CallTransferIndex = 0;
  m_ICTRequestPending = FALSE;
  m_IsIVRTransfer = FALSE;
  m_IVRDestroyConnectionOnReject = TRUE;
  
  ResourceCounter::IncrementConnection();
}

B2BUAConnection::~B2BUAConnection()
{
  m_CallSessions.AllowDeleteObjects();
  LOG_CONTEXT( LogInfo(), GetSessionId(), "B2BUAConnection Destroyed 0x" << std::hex << this );
  OStringStream rc;
  ResourceCounter::Print( rc );
  LOG_CONTEXT( LogInfo(), GetSessionId(), "*** COUNTERS *** " << rc.str() );
}

BOOL B2BUAConnection::AttachCall(
  const OString & id,
  B2BUACall * call,
  int index
)
{
  PWaitAndSignal lock( m_CallSessionsMutex );
  
  SIPSession::GCRef callRef = GCCREATEREF( call, "B2BUAConnection::AttachCall", "B2BUACall" );
  if( callRef == NULL )
    return FALSE;

  SIPSession::GCRef * cloneRef = new SIPSession::GCRef( callRef );

  if( m_CallSessions.Contains( id.c_str() ) )
  {
    SIPSession::GCRef * ref = m_CallSessions.RemoveAt( id.c_str() );
    delete ref;
  }

  m_CallSessions.SetAt( id.c_str(), cloneRef );

  if( index == -1 )
  {
    if( m_Leg1Call == NULL )
    {
      SetInterfacePort( call->GetInterfacePort() );
      m_Leg1Call = call;
    }else
    {
      m_Leg2Call = call;
    }
  }else if( index == 0 )
  {
    m_Leg1Call = call;
  }else if( index == 1 )
  {
    m_Leg2Call = call;
  }

  call->SetB2BUAConnection( this );

  return TRUE;
}

SIPSession::GCRef * B2BUAConnection::DetachCall(
  const OString & id
)
{
  PWaitAndSignal lock( m_CallSessionsMutex );
  return m_CallSessions.RemoveAt( id.c_str() );
}

SIPSession::GCRef * B2BUAConnection::FindCall(
  const OString & id
)
{
  PWaitAndSignal lock( m_CallSessionsMutex );
  return m_CallSessions.GetAt( id.c_str() );
}


BOOL B2BUAConnection::OnIncomingCall(
  B2BUACall & call,
  SIPMessage & invite
)
{
  B2BUserAgent & b2bua = 
    (B2BUserAgent &)GetSessionManager().GetUserAgent();

  return b2bua.OnIncomingCall( *this, call, invite );
}

void B2BUAConnection::OnRequireRedirectionId(
  B2BUACall & call,
  OString & redirectionId
)
{
  B2BUserAgent & b2bua = 
    (B2BUserAgent &)GetSessionManager().GetUserAgent();

  b2bua.OnRequireRedirectionId( *this, call, redirectionId );
}

void B2BUAConnection::OnRequireCallTransferId(
  B2BUACall & call,
  const ReferTo & referTo, 
  OString & redirectionId
)
{
  B2BUserAgent & b2bua = 
    (B2BUserAgent &)GetSessionManager().GetUserAgent();

  b2bua.OnRequireCallTransferId( *this, call, referTo, redirectionId );
}

BOOL B2BUAConnection::OnIncomingSDPOffer(
  B2BUACall & call,
  const SIPMessage & sdp
)
{
  B2BUserAgent & b2bua = 
    (B2BUserAgent &)GetSessionManager().GetUserAgent();
  return b2bua.OnIncomingSDPOffer( *this, call, sdp );
}

BOOL B2BUAConnection::OnIncomingSDPAnswer(
  B2BUACall & call,
  const SIPMessage & sdp
)
{
  if( !m_HasReceivedSDP )
    m_HasReceivedSDP = sdp.HasSDP();

  if( !m_HasReceivedSDP )
    return FALSE;
          
  B2BUserAgent & b2bua = 
    (B2BUserAgent &)GetSessionManager().GetUserAgent();

  return b2bua.OnIncomingSDPAnswer( *this, call, sdp );
}

/// your last chance to produce an answer to the offer.  
/// This will be called either before sending 183, 200 or ACK 
BOOL B2BUAConnection::OnRequireSDPAnswer(
  B2BUACall & call,
  const SIPMessage & offer,
  SIPMessage & answer
)
{
  B2BUserAgent & b2bua = 
    (B2BUserAgent &)GetSessionManager().GetUserAgent();
  return b2bua.OnRequireSDPAnswer( *this, call, offer, answer );
}

BOOL B2BUAConnection::OnSendB2BConnect(
  B2BUACall &,
  const SIPMessage & message
)
{
  B2BUserAgent & b2bua = 
    (B2BUserAgent &)GetSessionManager().GetUserAgent();
  return b2bua.OnSendB2BConnect( *this, *m_Leg2Call, message );
}

/// this will be called if an offer is not received in INVITE or when we are initiating the call
BOOL B2BUAConnection::OnRequireSDPOffer(
  B2BUACall & call,
  SIPMessage & offer
)
{
  B2BUserAgent & b2bua = 
    (B2BUserAgent &)GetSessionManager().GetUserAgent();
  return b2bua.OnRequireSDPOffer( *this, call, offer );
}

void B2BUAConnection::OnSetupOutbound( 
  const SIPMessage & inboundInvite 
)
{
  if( m_SessionState == Disconnected )
    return;

  B2BUserAgent & b2bua = 
    (B2BUserAgent &)GetSessionManager().GetUserAgent();

  SIPMessage outboundInvite = inboundInvite;
  outboundInvite.ClearInternalHeaders();

  StartAlertingTimer();
  StartSeizeTimer();
  b2bua.OnSetupOutbound( outboundInvite, *this );
}

void B2BUAConnection::OnSessionKeepAliveTimeout(
  B2BUACall & call 
)
{
  SIPMessage msg; /// dummy event
  DestroyConnection( msg );
}

/** NIST Callbacks */
void B2BUAConnection::NIST_OnReceivedRegister(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NIST_OnReceivedRefer(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NIST_OnReceivedBye(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NIST_OnReceivedOptions(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NIST_OnReceivedInfo(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NIST_OnReceivedMessage(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NIST_OnReceivedCancel(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NIST_OnReceivedNotify(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NIST_OnReceivedSubscribe(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NIST_OnReceivedUnknown(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}
void B2BUAConnection::IST_OnReceivedMoreInvite(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

/** NICT CALLBACKS */


void B2BUAConnection::NICT_OnReceived1xx(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NICT_OnReceived2xx(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NICT_OnReceived3xx(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NICT_OnReceived4xx(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NICT_OnReceived5xx(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NICT_OnReceived6xx(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::NICT_OnReceivedUnknown(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

/** ICT CALLBACKS */

void B2BUAConnection::ICT_OnReceived1xx(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::ICT_OnReceived2xx(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::ICT_OnReceived3xx(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::ICT_OnReceived4xx(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::ICT_OnReceived5xx(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::ICT_OnReceived6xx(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::ICT_OnReceivedUnknown(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  SetSessionState( call, msg );
}

void B2BUAConnection::SetSessionState(
  B2BUACall & call,
  const SIPMessage & msg
)
{
  PWaitAndSignal lock( m_SessionStateMutex );

  if( msg.IsInvite() )
    cout << "Got here" << endl;

  if( m_ConnectionDestroyed )
    return;
  
  if( m_SessionState < Connected && m_SessionState != Transferring && m_SessionState != AuthenticationPending )
  {
    OnHandlePreConnectState( call, msg );
  }else if( m_SessionState == Transferring )
  {
    OnHandleStateTransferring( call, msg );
  }else if( m_SessionState == AuthenticationPending )
  {
    OnHandleStateAuthenticationPending( call, msg );
  }else if( m_SessionState == Connected )
  {
    OnHandleStateConnected( call, msg );
  }else if( m_SessionState == Disconnected )
  {
    OnHandleStateDisconnected( call, msg );
  }else if( m_SessionState == Terminated )
  {
    OnHandleStateTerminated( call, msg );
  }else
  {
    PAssertAlways( PLogicError );
  }
}

/*
void B2BUAConnection::SetSessionState(
  B2BUACall & call,
  int event
)
{
  PWaitAndSignal lock( m_SessionStateMutex );
  
  if( m_SessionState == Idle )
  {
  }else if( m_SessionState == Trying )
  {
  }else if( m_SessionState == Proceeding )
  {
  }else if( m_SessionState == Connected )
  {
  }else if( m_SessionState == Disconnected )
  {
  }else if( m_SessionState == Terminated )
  {
  }else
  {
    PAssertAlways( PLogicError );
  }
}
*/

void B2BUAConnection::OnOutgoingCall(
  B2BUACall & call,
  SIPMessage & outboundInvite 
)
{
  B2BUserAgent& b2bUA = (B2BUserAgent&)GetSessionManager().GetUserAgent();
  b2bUA.OnOutgoingCall( *this, call, outboundInvite );
}

void B2BUAConnection::OnHandleLocalRefer( 
  B2BUACall & call,
  SIPMessage & refer 
)
{
  m_IsIVRTransfer = FALSE;
  B2BUserAgent& b2bUA = (B2BUserAgent&)GetSessionManager().GetUserAgent();
  b2bUA.OnLocalRefer( *this, call, refer );
  SIPMessage response;
  refer.CreateResponse( response, SIPMessage::Code202_Accepted );
  call.SendRequest( response );

  /// reset media flags
  m_HasReceivedSDP = FALSE;
  m_HasStartedMediaStream = FALSE;

  B2BMediaInterface::OpalMediaThread * audioThread = 
    (B2BMediaInterface::OpalMediaThread*)call.GetAudioRTPSessionThread();

  if( audioThread != NULL )
    audioThread->ForceExit();

  B2BMediaInterface::OpalMediaThread * videoThread = 
    (B2BMediaInterface::OpalMediaThread*)call.GetVideoRTPSessionThread();

  if( videoThread != NULL )
    videoThread->ForceExit();
  
  B2BUACall * otherEnd = NULL;
  OString sdp;
  if( call.GetLegIndex() == 0 )
  {
    otherEnd = m_Leg2Call;
    sdp = GetLeg2SDP();
    m_CallTransferIndex = 0;
  }else
  {
    otherEnd = m_Leg1Call;
    sdp = GetLeg1SDP();
    m_CallTransferIndex = 1;
  }

  /// set the session so that we get a special handler for REFER
  m_SessionState = Transferring;

  SIPMessage invite;

  /// Get the refer-to URI
  ReferTo referTo; 
  refer.GetReferTo( referTo );
  SIPURI requestURI = referTo.GetURI();

  /// create the request line
  RequestLine requestLine;
  requestLine.SetMethod( "INVITE" );
  requestLine.SetRequestURI( requestURI );
  invite.SetStartLine( requestLine );

  PIPSocket::Address targetAddress = requestURI.GetAddress();

  Via via;
  GetSessionManager().ConstructVia( targetAddress, via, SIPTransport::UDP );
  via.SetBranch( ParserTools::GenBranchParameter() );
  via.AddParameter( "rport", "" );
  invite.AppendVia( via );

  /// Set From header
  From from;
  from.SetURI( otherEnd->GetLocalURI() );
  from.AddParameter( "tag", ParserTools::GenTagParameter() );
  invite.SetFrom( from );

  /// Set To header
  To to;
  to.SetURI( requestURI );
  invite.SetTo( to );

  /// Set the call Id
  CallId callId;
  OString cid = GetLeg1Call()->GetCallId();
  callId.SetHeaderBody( cid );
  invite.SetCallId( callId );

  /// Set the CSeq
  CSeq cSeq;
  cSeq.SetMethod( "INVITE" );
  cSeq.SetSequence( 4711 );
  invite.SetCSeq( cSeq );

  /// Set the contact
  SIPURI contactURI = from.GetURI();
  contactURI.SetHost( via.GetURI().GetHost() );
  contactURI.SetPort( via.GetURI().GetPort() );
  ContactURI contact( contactURI, m_SessionProfile.GetDisplayName() );
  invite.AppendContact( contact );
  
  /// Set the allow header for Requests we support
  OStringArray allowed;
  allowed.AppendString( "INVITE" );
  allowed.AppendString( "BYE" );
  allowed.AppendString( "ACK" );
  allowed.AppendString( "REFER" );
  allowed.AppendString( "MESSAGE" );
  allowed.AppendString( "INFO" );
  allowed.AppendString( "NOTIFY" );
  allowed.AppendString( "OPTIONS" );
  Allow allow( allowed );
  invite.AppendAllow( allow );

  invite.SetSDP( sdp );

  RemoveRoutes();
  
  b2bUA.OnHandleLocalRefer( *this, call, invite );

  EnqueueSessionEvent( new SIPSessionEvent( *this, B2BUAConnection::SetupOutbound, invite ) );
}

void B2BUAConnection::OnHandlePreConnectState( 
  B2BUACall & call,
  const SIPMessage & message
)
{
  SIPMessage msg = message;

  B2BUserAgent& b2bUA = (B2BUserAgent&)GetSessionManager().GetUserAgent();

  if( call.GetLegIndex() == 1 )
  {
    if( msg.GetCSeqMethod() *= "INVITE" )
    {
      if( msg.Is1xx() )
      {
        if( msg.GetStatusCode() == 100 )
        {
          m_SessionState = Trying;
          return;
        }else
        {
          m_SessionState = Proceeding;
          // ok now we are ringing
          // stop connecting timer and start
          // alerting timer
          StopAlertingTimer();
          LOG( LogInfo(), "Connection: Alerting / SDP = " << ( msg.HasSDP() ? "TRUE" : "FALSE" ) );
          b2bUA.OnAlerting( *this, call, msg );
          OnIncomingSDPAnswer( call, msg );
        }
      }else if( msg.Is2xx() )
      {
        m_SessionState = Connected;
        // ok we receive 2xx value
        // stop alerting timer
        StopAlertingTimer();

        // we should also stop connection timer
        // due to it is possible that it
        // will directly send 200 without
        // sending 183 - alerting
        StopSeizeTimer(); 
        b2bUA.OnConnected( *this, call, msg );
        OnIncomingSDPAnswer( call, msg );

        // check if we would post a warning message
        if( !m_200OkWarning.GetHeaderBody().IsEmpty() )
        {
          // for now we would overrider warning messages
          // with our warning message
          msg.SetWarning( m_200OkWarning );
        }

        LOG( LogInfo(), "Connection: Connected / SDP = " << ( msg.HasSDP() ? "TRUE" : "FALSE" ) );
        /// Start the session keep alive timer

        if( m_SendSessionKeepAlive )
        {
          GetLeg1Call()->StartSessionKeepAlive( 120000 );
          GetLeg2Call()->StartSessionKeepAlive( 120000 );
        }

      }else if( !msg.IsRequest() ) // 3xx - 6xx
      {
        int statusCode = msg.GetStatusCode();
        // check if we should fail over
        if( !msg.Is3xx() && 
             statusCode != 401 &&
             statusCode != 407 && 
            !m_HasReceivedSDP )
        {
          if( m_Routes.GetSize() > 1 )
          {
            // destroy first leg 2 connection
            if( m_Leg2Call )
            {
              m_Leg2Call->Destroy();
              m_Leg2Call = NULL;
            }

            m_Routes.RemoveAt( 0 );

            LOG( LogInfo(), "Connection: Rejected / Code = " << msg.GetStatusCode() );

            // Change to FailOver method
            LOG( LogInfo(), "Connection: Failing over to other routes: Error " 
              << msg.GetStatusCode() );
            b2bUA.OnRejected( *this, call, msg );

            OnSetupOutbound( m_Leg1Call->GetRequest() );
            return;
          }
        } 

        // if message status is not 3xx then 
        // we should notify ua for rejection
        if( !msg.Is3xx() )
        {
          if(  statusCode != 401 && statusCode != 407 )
          {
            LOG( LogInfo(), "Connection: Rejected / Code = " << msg.GetStatusCode() );
            m_SessionState = Disconnected;
            b2bUA.OnRejected( *this, call, msg );
          }else
          {
            m_SessionState = AuthenticationPending;
            /// Stop the application timers to make sure they dont fire when we are in AuthenticationPending state
            StopSeizeTimer();
            StopAlertingTimer();
            /// start the auto destructor timer at the value of Timer B
            StartAutoDestructTimer( 32000 );
            SIPMessage challenge;
            if( statusCode == 401 )
            {
              GetLeg1Call()->GetCurrentUASInvite().CreateResponse( challenge, SIPMessage::Code401_Unauthorized, TRUE );
              WWWAuthenticate auth;
              if( msg.GetWWWAuthenticate( auth ) )
                challenge.SetWWWAuthenticate( auth );
            }else if( statusCode == 407 )
            {
              GetLeg1Call()->GetCurrentUASInvite().CreateResponse( challenge, SIPMessage::Code407_ProxyAuthenticationRequired, TRUE );
              ProxyAuthenticate auth;
              if( msg.GetProxyAuthenticate( auth ) )
                challenge.SetProxyAuthenticate( auth );
            }

            GetLeg1Call()->SendRequest( challenge );

            return;
          }
        }else
        {
          /// just return as if nothing happend, call is now about to be redirected
          
          return;
        }
      }

      GetLeg1Call()->AnswerB2BCall( call, msg );
    }
  }else if( call.GetLegIndex() == 0 )
  {
    if( msg.IsCancel() )
    {
      SIPMessage ok;
      msg.CreateResponse( ok, SIPMessage::Code200_Ok );
      call.SendRequest( ok );
      SIPMessage reqCancelled;
      call.GetRequest().CreateResponse( reqCancelled, SIPMessage::Code487_RequestCancelled );
      call.SendRequest( reqCancelled );
      
      if( m_SessionState < Connected )
      {
        if( m_EnableForkCalls && m_ForkCalls )
        {
          m_ForkCalls->CancelForkCalls();
        } else 
        {
          B2BUACall * leg2Call = GetLeg2Call();
          if( leg2Call )
            leg2Call->SendCancel();
        }
      }

      m_SessionState = Disconnected;
      LOG( LogInfo(), "Connection: Abandoned" );
      b2bUA.OnCallAbandoned( *this, call, msg );
    }     
  }else
  {
    PAssertAlways( PLogicError );
  }

  if( m_SessionState == Disconnected )
  {
    // if we receive disconnect then
    // stop all timer
    StopSeizeTimer();
    StopAlertingTimer();
    DestroyConnection( msg );    
  } 
}

void B2BUAConnection::OnHandleStateTransferring( 
  B2BUACall & call,
  const SIPMessage & message
)
{
  /// ignore events from the referrer call
  if( call.IsReferrerCall() )
  {
    LOG( LogWarning(), "OnHandleStateTransferring - Ignoring " << message.GetMethod() );
    return;
  }

  LOG_IF_DEBUG( LogDebugVeryHigh(), message );

  B2BUserAgent& b2bUA = (B2BUserAgent&)GetSessionManager().GetUserAgent();
  SIPMessage msg = message;
  if( !call.IsReferredCall() )
  {
    if( msg.IsBye() )
    {
      if( GetLeg1Call()->IsReferredCall() )
      {
        GetLeg1Call()->SendCancel();
        GetLeg1Call()->ReleaseTransferedCall();
      }else if( GetLeg2Call()->IsReferredCall() )
      {
        GetLeg2Call()->SendCancel();
        GetLeg2Call()->ReleaseTransferedCall();
      }

      m_SessionState = Disconnected;
      LOG( LogInfo(), "Connection: Abandoned" );
      b2bUA.OnCallAbandoned( *this, call, msg );
      StopSeizeTimer();
      StopAlertingTimer();
      DestroyConnection( msg );  
    }

    return;
  }

  

  if( msg.GetCSeqMethod() *= "INVITE" )
  {
    if( msg.Is1xx() )
    {
      if( msg.GetStatusCode() == 100 )
      {
        return;
      }else
      {
        // ok now we are ringing
        // stop connecting timer and start
        // alerting timer
        StopAlertingTimer();
        LOG( LogInfo(), "Connection: Alerting / SDP = " << ( msg.HasSDP() ? "TRUE" : "FALSE" ) );
        b2bUA.OnAlerting( *this, call, msg );

        OnIncomingSDPAnswer( call, msg );

        if( m_HasReceivedSDP )
        {
          SIPMessage answer;
          OnRequireSDPAnswer( call, msg, answer );
        }
      }
    }else if( msg.Is2xx() )
    {
      m_SessionState = Connected;
      // ok we receive 2xx value
      // stop alerting timer
      StopAlertingTimer();

      // we should also stop connection timer
      // due to it is possible that it
      // will directly send 200 without
      // sending 183 - alerting
      StopSeizeTimer(); 
      b2bUA.OnConnected( *this, call, msg );

      // check if we would post a warning message
      if( !m_200OkWarning.GetHeaderBody().IsEmpty() )
      {
        // for now we would overrider warning messages
        // with our warning message
        msg.SetWarning( m_200OkWarning );
      }

      LOG( LogInfo(), "Connection: Connected / SDP = " << ( msg.HasSDP() ? "TRUE" : "FALSE" ) );
      /// Start the session keep alive timer

      if( m_SendSessionKeepAlive )
      {
        GetLeg1Call()->StartSessionKeepAlive( 120000 );
        GetLeg2Call()->StartSessionKeepAlive( 120000 );
      }

      OnIncomingSDPAnswer( call, msg );

      if( m_HasReceivedSDP )
      {
        SIPMessage answer;
        OnRequireSDPAnswer( call, msg, answer );
      }

      call.ReleaseTransferedCall();
      call.SetReferredCall( FALSE );

    }else
    {
      // if we receive disconnect then
      // stop all timer
      StopSeizeTimer();
      StopAlertingTimer();
      call.ReleaseTransferedCall();
      call.SetReferredCall( FALSE );

	  if( m_IsIVRTransfer && m_IVRDestroyConnectionOnReject )
		  DestroyConnection( msg );
    else if( m_IsIVRTransfer && !m_IVRDestroyConnectionOnReject )
      IVROnTransferReject( msg );
    else
      DestroyConnection( msg );
    }
  }
}

void B2BUAConnection::OnHandleStateAuthenticationPending( 
  B2BUACall & call,
  const SIPMessage & msg
)
{
  if( msg.IsInvite() )
  {
    LOG( LogInfo(), "Got INVITE resend for AuthenticationPending" );
    StopAutoDestructTimer();
    /// Get a reference to the leg 2 call
    SIPSession::GCRef gcref = GCCREATEREF( GetLeg2Call(), "B2BUAConnection::OnHandleStateAuthenticationPending", "CallSession" );
    if( gcref == NULL )
    {
      SIPMessage response;
      msg.CreateResponse( response, SIPMessage::Code481_TransactionDoesNotExist );
      DestroyConnection(response);
      return;
    }
    
    SIPMessage oldUACInvite = GetLeg2Call()->GetCurrentUACInvite();
    if( !oldUACInvite.IsValid() )
    {
      SIPMessage response;
      msg.CreateResponse( response, SIPMessage::Code481_TransactionDoesNotExist );
      DestroyConnection(response);
      return;
    }
    
    if( msg.HasProxyAuthorization() )
    {
      ProxyAuthorization auth;
      if( msg.GetProxyAuthorization( auth ) )
        oldUACInvite.SetProxyAuthorization( auth );
      else
      {
        SIPMessage response;
        msg.CreateResponse( response, SIPMessage::Code481_TransactionDoesNotExist );
        DestroyConnection(response);
        return;
      }
    }else if( msg.HasAuthorization() )
    {
      Authorization auth;
      if( msg.GetAuthorization( auth ) )
        oldUACInvite.SetAuthorization( auth );
      else
      {
        SIPMessage response;
        msg.CreateResponse( response, SIPMessage::Code481_TransactionDoesNotExist );
        DestroyConnection(response);
        return;
      }
    }else
    {
      SIPMessage response;
      msg.CreateResponse( response, SIPMessage::Code481_TransactionDoesNotExist );
      DestroyConnection(response);
      return;
    }
    
    /// revert the state to trying
    m_SessionState = Trying;
    
    /// restart the call timers
    StartSeizeTimer();
    StartAlertingTimer();
    
    /// we got through the if bomabardments
    /// lets increment the CSeq before sending the invite outbound
    oldUACInvite.IncrementCSeq();

    /// Lets change the via branch to ensure we don't collide with the old transaction
    Via via = oldUACInvite.PopTopVia();
    via.AddParameter( "branch", ParserTools::GenBranchParameter(), TRUE );
    oldUACInvite.AppendVia( via );

    LOG( LogDebugVeryHigh(), oldUACInvite );
    GetLeg2Call()->SetCurrentUACInvite( oldUACInvite );
    GetLeg2Call()->SendRequest( oldUACInvite );
    return;
  }
   
}

void B2BUAConnection::OnHandleStateConnected( 
  B2BUACall & call,
  const SIPMessage & msg
)
{
  /// ignore messages from the referrer leg
  if( call.IsReferrerCall() )
    return;

  B2BUACall * otherEnd = NULL;
  if( call.GetLegIndex() == 0 )
    otherEnd = GetLeg2Call();
  else
    otherEnd = GetLeg1Call();


  if( msg.IsCancel() )
  { // cancel is not okayed automatically
    // by the CallSession so we need to ok it here
    // since we are already in connected state,
    // Cancel is already late so just do nothing but respond
    SIPMessage ok;
    msg.CreateResponse( ok, SIPMessage::Code200_Ok );
    call.SendRequest( ok );
  }else if( msg.IsBye() )
  {
    otherEnd->SendBye();
    m_SessionState = Disconnected;
  }else if( msg.IsInfo() || msg.IsMessage() )
  {
    SIPMessage ok;
    msg.CreateResponse( ok, SIPMessage::Code200_Ok );
    call.SendRequest( ok );
    
    SIPMessage clone;
    if( CloneRequestWithinDialog( msg, clone ) )
      otherEnd->SendRequest( clone );

  }else if( msg.IsInvite() && msg.HasSDP() )
  {
    if( m_ICTRequestPending )
    {
      SIPMessage requestPending;
      msg.CreateResponse( requestPending, SIPMessage::Code491_RequestPending );
      call.SendRequest( requestPending );
    }else
    {
      m_ICTRequestPending = TRUE;
      otherEnd->SendReinvite();
    }
  }else if( msg.IsRefer() )
  {
    if( m_EnableLocalRefer )
    {
      SIPMessage refer = msg;
      OnHandleLocalRefer( call, refer ); 
    } else 
    {
      SIPMessage xferResponse;
      if( !otherEnd->SendB2BRefer( call, msg ) )
      {
        msg.CreateResponse( xferResponse, SIPMessage::Code406_NotAcceptable );
        call.SendRequest( xferResponse );
      } else
      {
        msg.CreateResponse( xferResponse, SIPMessage::Code202_Accepted );
        SendRequest( xferResponse );
        StartAutoDestructTimer( 90000 );
      }
    } 
  }else if( msg.IsNotify() )
  {
    SIPMessage notifyResponse;
    if( !otherEnd->SendB2BNotify( call, msg ) )
    {
      msg.CreateResponse( notifyResponse, SIPMessage::Code406_NotAcceptable );
      call.SendRequest( notifyResponse );
    }
  }else if( !msg.IsRequest() )
  {
    if( msg.GetCSeqMethod() *= "INVITE" )
    {
      m_ICTRequestPending = FALSE;
      /*
      RTP_SessionManager & rtpManager = call.GetRTPSessionManager();
      RTP_UDP * rtp = (RTP_UDP *)rtpManager.GetSession( 
      OpalMediaFormat::DefaultAudioSessionID );
      if( rtp == NULL )*/
      if( msg.Is2xx() )
        otherEnd->SendB2BConnect( msg );
      else if( msg.Is3xx() )
        otherEnd->SendB2BRedirect( msg );
      else if( msg.Is4xx() || msg.Is5xx() || msg.Is6xx() )
        otherEnd->SendB2BReject( msg );
    }
  }

  if( m_SessionState == Disconnected )
  {
    B2BUserAgent& b2bUA = (B2BUserAgent&)GetSessionManager().GetUserAgent();
    b2bUA.OnDisconnected( *this, call, msg );
    LOG( LogInfo(), "Connection: Disconnected" );
    DestroyConnection( msg );
  }  
}

void B2BUAConnection::OnHandleStateDisconnected( 
  B2BUACall & call,
  const SIPMessage & msg
)
{
  /// ignore messages from the referrer leg
  if( call.IsReferrerCall() )
    return;

  if( msg.IsRequest() )
  {
    SIPMessage response;
    msg.CreateResponse( response, SIPMessage::Code481_TransactionDoesNotExist );
    call.SendRequest( response );
  }
}

void B2BUAConnection::OnHandleStateTerminated( 
  B2BUACall & call,
  const SIPMessage & msg
)
{
  /// ignore messages from the referrer leg
  if( call.IsReferrerCall() )
    return;

  if( msg.IsRequest() )
  {
    SIPMessage response;
    msg.CreateResponse( response, SIPMessage::Code481_TransactionDoesNotExist );
    call.SendRequest( response );
  }
}

void B2BUAConnection::OnSessionEvent(
  int event,
  const SIPMessage & eventMsg
)
{
  switch( event )
  {
    case SetupOutbound:
      OnSetupOutbound( eventMsg );
      break;
    case DestroySession:
      SetGCLifeSpan( 32 ); /// Lifespan set to 64 * T1
      Destroy();
      break;
    case ErrorAlertingTimeout:
    case ErrorSeizeTimeout:
      { 
        PWaitAndSignal lock( m_SessionStateMutex );

        CallSession::GCRef leg2Ref = GCCREATEREF( m_Leg2Call, "B2BUAConnection::OnSessionEvent", "B2BUACall Leg 2" ); 

        if( leg2Ref != NULL )
        {
          if( m_SessionState <= Proceeding )
          {
            B2BUACall * leg2Call = static_cast<B2BUACall*>(leg2Ref.GetObject());
            SIPMessage timeout;
            leg2Call->GetCurrentUACRequest().CreateResponse( timeout, 
              SIPMessage::Code408_RequestTimeout );

            OnHandlePreConnectState( *leg2Call, timeout );
          }
        }
      }
      break;
    case ErrorNoRoute:
      Destroy();
      break;
    case ErrorRouteResolution:
      Destroy();
      break;
  }
}

void B2BUAConnection::DestroyConnection()
{
  SIPMessage dummy;
  DestroyConnection( dummy );
}

void B2BUAConnection::DestroyConnection(
  const SIPMessage & msg
)
{
  EnqueueSessionEvent( new SIPSessionEvent( *this, DestroySession, msg ) );
}

#if 0
void B2BUAConnection::OnDestroySession()
{
  PWaitAndSignal lock( m_SessionStateMutex );

  if( m_ConnectionDestroyed )
    return;

  m_ConnectionDestroyed = TRUE;
  
  // Required data for Sending Auth
  // required before leg 1 is destroyed
  SIPMessage previousInvite = m_Leg1Call->GetRequest();
  ProxyAuthenticate auth = m_Leg1Call->GetProxyAuthenticate();

  if( m_SessionState == Connected )
  {
    m_Leg1Call->SendBye();
    if( m_Leg2Call != NULL )
      m_Leg2Call->SendBye();
  }else if( m_SessionState < Connected )
  {
    if( m_Leg1Call->GetCallAnswerResponse() 
      != CallSession::DisconnectWithUnauthorized && 
      m_Leg1Call->GetCallAnswerResponse() != CallSession::DontDeny
    )
    {
      SIPMessage reject;
      previousInvite.CreateResponse( reject, (SIPMessage::StatusCodes)m_Leg1Call->GetCallAnswerResponse() );
      SendRequest( reject );

      if( m_EnableForkCalls )
      {
        m_ForkCalls->CancelForkCalls();
      }
      else if( m_Leg2Call != NULL )
      {
        m_Leg2Call->SendCancel();
      }
    }
  }

  m_Leg1Call->CloseMediaStreams();
  if( m_Leg2Call != NULL )
    m_Leg2Call->CloseMediaStreams();

  m_Leg1Call->Destroy();
  if( m_Leg2Call != NULL )
    m_Leg2Call->Destroy();

  // Handle here for sending Proxy Authentication Required
  // Problem: Causes was Race Condition
  // Todo: Should be fix race condition and put
  // this method inside on B2BUA::OnIncomingCall
  SIPMessage * authChallenge = NULL;
  if( m_Leg1Call->GetCallAnswerResponse() 
    == CallSession::DisconnectWithUnauthorized )
  {
    authChallenge = new SIPMessage();
    previousInvite.CreateResponse( *authChallenge, 
      SIPMessage::Code407_ProxyAuthenticationRequired );
    
    authChallenge->SetProxyAuthenticate( auth );
  }

  if( m_ForkCalls )
  {
    delete m_ForkCalls;
    m_ForkCalls = NULL;
  }

  PWaitAndSignal lock2( m_CallSessionsMutex );
  m_CallSessions.AllowDeleteObjects();
  m_CallSessions.RemoveAll();

  m_Leg1Call = NULL;
  m_Leg2Call = NULL;

  if( authChallenge != NULL )
    GetSessionManager().EnqueueSessionManagerEvent( authChallenge );
 
}

void B2BUAConnection::OnGCCollect()
{
}

#else
void B2BUAConnection::OnDestroySession()
{
  PWaitAndSignal lock( m_SessionStateMutex );

  if( m_ConnectionDestroyed )
    return;

  m_ConnectionDestroyed = TRUE;
  
  // Required data for Sending Auth
  // required before leg 1 is destroyed
  SIPMessage previousInvite = m_Leg1Call->GetRequest();
  ProxyAuthenticate auth = m_Leg1Call->GetProxyAuthenticate();

  if( m_SessionState == Connected )
  {
    m_Leg1Call->SendBye();
    if( m_Leg2Call != NULL )
      m_Leg2Call->SendBye();
  }else if( m_SessionState < Connected )
  {
    if( m_Leg1Call->GetCallAnswerResponse() 
      != CallSession::DisconnectWithUnauthorized && 
      m_Leg1Call->GetCallAnswerResponse() != CallSession::DontDeny
    )
    {
      SIPMessage reject;
      previousInvite.CreateResponse( reject, (SIPMessage::StatusCodes)m_Leg1Call->GetCallAnswerResponse() );
      SendRequest( reject );

      if( m_EnableForkCalls )
      {
        m_ForkCalls->CancelForkCalls();
      }
      else if( m_Leg2Call != NULL )
      {
        m_Leg2Call->SendCancel();
      }
    }
  }

  m_Leg1Call->QueueForDestruction();
  if( m_Leg2Call != NULL )
    m_Leg2Call->QueueForDestruction();

  // Handle here for sending Proxy Authentication Required
  // Problem: Causes was Race Condition
  // Todo: Should be fix race condition and put
  // this method inside on B2BUA::OnIncomingCall
  SIPMessage * authChallenge = NULL;
  if( m_Leg1Call->GetCallAnswerResponse() 
    == CallSession::DisconnectWithUnauthorized )
  {
    authChallenge = new SIPMessage();
    previousInvite.CreateResponse( *authChallenge, 
      SIPMessage::Code407_ProxyAuthenticationRequired );
    
    authChallenge->SetProxyAuthenticate( auth );
    SetGCLifeSpan(0);
  }

  if( authChallenge != NULL )
    GetSessionManager().EnqueueSessionManagerEvent( authChallenge );
  
  ResourceCounter::DecrementConnection();
 
}

void B2BUAConnection::OnSelfDestruct()
{
  DestroyConnection();
}

void B2BUAConnection::OnGCCollect()
{
  m_Leg1Call->CloseMediaStreams();
  if( m_Leg2Call != NULL )
    m_Leg2Call->CloseMediaStreams();

  m_Leg1Call->PurgeDestructionQueue();
  if( m_Leg2Call != NULL )
    m_Leg2Call->PurgeDestructionQueue();

  if( m_ForkCalls )
  {
    delete m_ForkCalls;
    m_ForkCalls = NULL;
  }

  PWaitAndSignal lock2( m_CallSessionsMutex );
  m_CallSessions.AllowDeleteObjects();
  m_CallSessions.RemoveAll();

  m_Leg1Call = NULL;
  m_Leg2Call = NULL;
}

#endif
void B2BUAConnection::StartSeizeTimer()
{
  if( m_SeizeTimeout > 0 )
  {
    m_SeizeTimer = PTimer( m_SeizeTimeout );
    
    m_SeizeTimer.SetNotifier( PCREATE_NOTIFIER( OnSeizeTimeout ) );
    m_SeizeTimer.Resume();
  }
}

void B2BUAConnection::StopSeizeTimer()
{
  m_SeizeTimer.Stop();
}

void B2BUAConnection::StartAlertingTimer()
{
  if( m_AlertingTimeout > 0 )
  {
    m_AlertingTimer = PTimer( m_AlertingTimeout );
    
    m_AlertingTimer.SetNotifier( PCREATE_NOTIFIER( OnAlertingTimeout ) );
    m_AlertingTimer.Resume();
  }
}

void B2BUAConnection::StopAlertingTimer()
{
  m_AlertingTimer.Stop();
}


void B2BUAConnection::OnSeizeTimeout( 
  PTimer&,
  INT
)
{
  EnqueueSessionEvent( new SIPSessionEvent( *this, B2BUAConnection::ErrorSeizeTimeout ) );
}

void B2BUAConnection::OnAlertingTimeout( 
  PTimer&,
  INT 
)
{
  EnqueueSessionEvent( new SIPSessionEvent( *this, B2BUAConnection::ErrorAlertingTimeout ) );
}

OString B2BUAConnection::GenerateCallIdExtension()
{
  static PInt16 counter = 0;

  OStringStream stream;
  stream << "0x" << std::setw(4) << std::setfill('0') << std::hex << ++counter;
  return stream.str();
}

/// IVR related members
BOOL B2BUAConnection::AttachIVRContext(
  IVRContext * context
)
{
  PWaitAndSignal lock( m_IVRContextMutex );
  m_IVRContext = context;
  m_IVRContext->AttachB2BUAConnection( this );
  OnAttachIVRContext( context );
  return TRUE;
}

void B2BUAConnection::OnAttachIVRContext( 
  IVRContext * context 
)
{
}

BOOL B2BUAConnection::DetachIVRContext()
{
  PWaitAndSignal lock( m_IVRContextMutex );
  OnDetachIVRContext();
  m_IVRContext = NULL;
  return TRUE;
}

void B2BUAConnection::OnDetachIVRContext()
{
}

BOOL B2BUAConnection::IVRPlayText(
  const OString & text 
)
{
  PWaitAndSignal lock( m_IVRContextMutex );
  if( m_IVRContext == NULL )
    return FALSE;
  return m_IVRContext->PlayText( text, PTextToSpeech::Default, 0, 0 );
}

BOOL B2BUAConnection::IVRPlayFile(
  const OString & id,
  const OString & fn 
)
{
  PWaitAndSignal lock( m_IVRContextMutex );
  if( m_IVRContext == NULL )
    return FALSE;
  return m_IVRContext->PlayFile( id, fn, 0, 0, FALSE );
}

BOOL B2BUAConnection::IVRPlayData(
  const OString & id,
  const PBYTEArray & data
)
{
  PWaitAndSignal lock( m_IVRContextMutex );
  
  if( m_IVRContext == NULL )
    return FALSE;

  return m_IVRContext->PlayData( id, data, 0, 0 );
}

BOOL B2BUAConnection::IVRPlayResource(
  const OString & id,
  const PURL & url
)
{
  PWaitAndSignal lock( m_IVRContextMutex );
  
  if( m_IVRContext == NULL )
    return FALSE;

  return m_IVRContext->PlayResource( id, url, 0, 0 );
}

BOOL B2BUAConnection::IVRPlaySilence(
  PINDEX msecs
)
{
  PWaitAndSignal lock( m_IVRContextMutex );
  
  if( m_IVRContext == NULL )
    return FALSE;

  return m_IVRContext->PlaySilence( msecs );
}

BOOL B2BUAConnection::IVRPlaySilence(
  const PTimeInterval & timeout
)
{
  PWaitAndSignal lock( m_IVRContextMutex );
  
  if( m_IVRContext == NULL )
    return FALSE;

  return m_IVRContext->PlaySilence( timeout );
}

BOOL B2BUAConnection::IVRStartRecording(
  const PFilePath & fn, 
  BOOL recordDTMFTerm, 
  const PTimeInterval & recordMaxTime, 
  const PTimeInterval & recordFinalSilence
)
{
  PWaitAndSignal lock( m_IVRContextMutex );
  
  if( m_IVRContext == NULL )
    return FALSE;

  return m_IVRContext->StartRecording( fn, recordDTMFTerm, recordMaxTime, recordFinalSilence  );
}

BOOL B2BUAConnection::IVREndRecording()
{
  PWaitAndSignal lock( m_IVRContextMutex );
  
  if( m_IVRContext == NULL )
    return FALSE;

  return m_IVRContext->EndRecording();
}

BOOL B2BUAConnection::IVRIsPlaying() const
{
  PWaitAndSignal lock( m_IVRContextMutex );
  
  if( m_IVRContext == NULL )
    return FALSE;

  return m_IVRContext->IsPlaying();
}

BOOL B2BUAConnection::IVRIsRecording() const
{
  PWaitAndSignal lock( m_IVRContextMutex );
  
  if( m_IVRContext == NULL )
    return FALSE;

  return m_IVRContext->IsPlaying();
}

PTextToSpeech * B2BUAConnection::IVRSetTextToSpeech(
  PTextToSpeech * _tts, 
  BOOL autoDelete
)
{
  PWaitAndSignal lock( m_IVRContextMutex );
  
  if( m_IVRContext == NULL )
    return FALSE;

  return m_IVRContext->SetTextToSpeech( _tts, autoDelete );
}

PTextToSpeech * B2BUAConnection::IVRSetTextToSpeech(
  const OString & ttsName
)
{
  PWaitAndSignal lock( m_IVRContextMutex );
  
  if( m_IVRContext == NULL )
    return FALSE;

  return m_IVRContext->SetTextToSpeech( ttsName );
}

PTextToSpeech * B2BUAConnection::IVRGetTextToSpeech()
{
  PWaitAndSignal lock( m_IVRContextMutex );
  
  if( m_IVRContext == NULL )
    return FALSE;

  return m_IVRContext->GetTextToSpeech();
}

BOOL B2BUAConnection::IVRSetDTMFGrammar(
    const OString & identifier,
    int maxDigits, /// maximum digitd collected
    char termChar, /// terminating char to end collection
    BOOL canStopPlayBack,
    const PTimeInterval & collectTimeout,
    const PTimeInterval & perDigitTimeout
)
{
  PWaitAndSignal lock( m_IVRContextMutex );
  
  if( m_IVRContext == NULL )
    return FALSE;

  return m_IVRContext->SetDTMFGrammar( identifier, maxDigits, termChar, canStopPlayBack, collectTimeout, perDigitTimeout );
}

void B2BUAConnection::IVROnEndPlaying( 
  const OString & identifier 
)
{
  GCVERIFYREF_VOID( "B2BUAConnection::IVROnEndPlaying" );
  PTRACE( 1, "Stopped playing " << identifier );
  B2BUserAgent & b2bua = 
    (B2BUserAgent &)GetSessionManager().GetUserAgent();
  b2bua.IVROnEndPlaying( *this, identifier );
}

BOOL B2BUAConnection::IVROnUserInput(
  const OString & str
)
{
  GCVERIFYREF( "B2BUAConnection::IVROnUserInput", FALSE ) ;
  PTRACE( 1, "User Input: " << str );
  B2BUserAgent & b2bua = 
    (B2BUserAgent &)GetSessionManager().GetUserAgent();
  return b2bua.IVROnUserInput( *this, str );
}

BOOL B2BUAConnection::IVROnOpenChannel(
  IVRContext * context 
)
{
  GCVERIFYREF( "B2BUAConnection::IVROnOpenChannel", FALSE ) ;
  B2BUserAgent & b2bua = 
    (B2BUserAgent &)GetSessionManager().GetUserAgent();
  return b2bua.IVROnOpenChannel( *this );
}

void B2BUAConnection::IVROnEndCollection( 
  const OString &grammarId, 
  const OString &grammarBuffer 
)
{
  GCVERIFYREF_VOID( "B2BUAConnection::IVROnOpenChannel" ) ;
  B2BUserAgent & b2bua = 
    (B2BUserAgent &)GetSessionManager().GetUserAgent();
  b2bua.IVROnEndCollection( *this, grammarId, grammarBuffer );
}

void B2BUAConnection::IVROnTransferReject(
  const SIPMessage & reject
)
{
  GCVERIFYREF_VOID( "B2BUAConnection::IVROnTransferReject" ) ;
  B2BUserAgent & b2bua = 
    (B2BUserAgent &)GetSessionManager().GetUserAgent();
  b2bua.IVROnTransferReject( *this, reject );
}

BOOL B2BUAConnection::IVRTransferCall(
  const OString & destNumber,
  BOOL destroyConnectionOnReject
)
{
  GCVERIFYREF( "B2BUAConnection::IVRTransferCall", FALSE ) ;

  SIPURI destURI;
  if( destNumber.Find( "@" ) == P_MAX_INDEX )
  {
    destURI = GetLeg1Call()->GetCurrentUASInvite().GetToURI();
    destURI.SetUser( destNumber );
  }else
  {
    destURI = destNumber;
  }

  return IVRTransferCall( destURI, destroyConnectionOnReject );

}

BOOL B2BUAConnection::IVRTransferCall(
  SIPURI & requestURI,
  BOOL destroyConnectionOnReject
)
{
  GCVERIFYREF( "B2BUAConnection::IVRTransferCall", FALSE ) ;
 
  m_IVRDestroyConnectionOnReject = destroyConnectionOnReject;
  m_IsIVRTransfer = TRUE;

  B2BUserAgent& b2bUA = (B2BUserAgent&)GetSessionManager().GetUserAgent();
  B2BUACall * call = GetLeg2Call();

  /// reset media flags
  m_HasReceivedSDP = FALSE;
  m_HasStartedMediaStream = FALSE;

  B2BMediaInterface::OpalMediaThread * audioThread = 
    (B2BMediaInterface::OpalMediaThread*)call->GetAudioRTPSessionThread();

  if( audioThread != NULL )
    audioThread->ForceExit();

  B2BMediaInterface::OpalMediaThread * videoThread = 
    (B2BMediaInterface::OpalMediaThread*)call->GetVideoRTPSessionThread();

  if( videoThread != NULL )
    videoThread->ForceExit();
  
  B2BUACall * otherEnd = m_Leg1Call;
  OString sdp = GetLeg1SDP();
  m_CallTransferIndex = 1;


  /// set the session so that we get a special handler for REFER
  m_SessionState = Transferring;

  SIPMessage invite;

  /// create the request line
  RequestLine requestLine;
  requestLine.SetMethod( "INVITE" );
  requestLine.SetRequestURI( requestURI );
  invite.SetStartLine( requestLine );

  PIPSocket::Address targetAddress = requestURI.GetAddress();

  Via via;
  GetSessionManager().ConstructVia( targetAddress, via, SIPTransport::UDP );
  via.SetBranch( ParserTools::GenBranchParameter() );
  via.AddParameter( "rport", "" );
  invite.AppendVia( via );

  /// Set From header
  From from;
  from.SetURI( otherEnd->GetLocalURI() );
  from.AddParameter( "tag", ParserTools::GenTagParameter() );
  invite.SetFrom( from );

  /// Set To header
  To to;
  to.SetURI( requestURI );
  invite.SetTo( to );

  /// Set the call Id
  CallId callId;
  OString cid = GetLeg1Call()->GetCallId();
  callId.SetHeaderBody( cid );
  invite.SetCallId( callId );

  /// Set the CSeq
  CSeq cSeq;
  cSeq.SetMethod( "INVITE" );
  cSeq.SetSequence( 4711 );
  invite.SetCSeq( cSeq );

  /// Set the contact
  SIPURI contactURI = from.GetURI();
  contactURI.SetHost( via.GetURI().GetHost() );
  contactURI.SetPort( via.GetURI().GetPort() );
  ContactURI contact( contactURI, m_SessionProfile.GetDisplayName() );
  invite.AppendContact( contact );
  
  /// Set the allow header for Requests we support
  OStringArray allowed;
  allowed.AppendString( "INVITE" );
  allowed.AppendString( "BYE" );
  allowed.AppendString( "ACK" );
  allowed.AppendString( "REFER" );
  allowed.AppendString( "MESSAGE" );
  allowed.AppendString( "INFO" );
  allowed.AppendString( "NOTIFY" );
  allowed.AppendString( "OPTIONS" );
  Allow allow( allowed );
  invite.AppendAllow( allow );

  invite.SetSDP( sdp );

  RemoveRoutes();
  
  b2bUA.OnHandleLocalRefer( *this, *call, invite );


  EnqueueSessionEvent( new SIPSessionEvent( *this, B2BUAConnection::SetupOutbound, invite ) );

  return TRUE;
}


