
/*
 *
 * OString.cxx
 * 
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: OString.cxx,v $
 * Revision 1.4  2007/07/16 09:43:56  joegenbaclor
 * PTRACE now uses internal OSS Logger
 *
 * Revision 1.3  2007/07/15 07:50:04  joegenbaclor
 * tidying up resource counters
 *
 * Revision 1.1  2007/07/14 07:50:34  joegenbaclor
 * Migrated code to use new STL based string class
 *
 *
 */

#include "OString.h"
using namespace Tools;

template <class T> char * ext_unsigned2string(T value, T base, char * str)
{
  if (value >= base)
    str = ext_unsigned2string<T>(value/base, base, str);
  value %= base;
  if (value < 10)
    *str = (char)(value + '0');
  else
    *str = (char)(value + 'A'-10);
  return str+1;
}


template <class T> char * ext_signed2string(T value, T base, char * str)
{
  if (value >= 0)
    return ext_unsigned2string<T>(value, base, str);

  *str = '-';
  return ext_unsigned2string<T>(-value, base, str+1);
}

OString::OString() : ext_string() 
{ 
}

OString::OString( char c )  : ext_string()
{
  string::resize(1);
  (*this)[0] = c;
}

OString::OString(
  const PString & s
)  : ext_string( (const char *)s )
{
}

OString::OString(
  const PCaselessString & s
)  : ext_string( (const char *)s )
{
}

OString::OString(
  const ostringstream & s
  )  : ext_string( s.str() )
{
}


OString::OString(
  const string &s, 
  size_type pos, 
  size_type n
) : ext_string(s, pos, npos) 
{ 
}

OString::OString(
  const value_type *s
) : ext_string(s) { }

OString::OString(
  const value_type *s, 
  size_type n
) : ext_string(s, n) { }


OString::OString(
  size_type n, 
  value_type c
) : ext_string(n, c) 
{ 
}

OString::OString(short n) : ext_string()
{
  char * theArray = (char *)calloc(sizeof(short)*3+1, sizeof(char));
  ext_signed2string<int>(n, 10, theArray);
  *this = (const char *)theArray;
  free( theArray );
}


OString::OString(unsigned short n) : ext_string()
{
  char * theArray = (char *)calloc(sizeof(unsigned short)*3+1, sizeof(char));
  ext_unsigned2string<unsigned short>(n, 10, theArray);
  *this = (const char *)theArray;
  free( theArray );
}


OString::OString(int n) : ext_string()
{
  char * theArray = (char *)calloc(sizeof(int)*3+1, sizeof(char));
  ext_signed2string<int>(n, 10, theArray);
  *this = (const char *)theArray;
  free( theArray );
}


OString::OString(unsigned int n) : ext_string()
{
  char * theArray = (char *)calloc(sizeof(unsigned int)*3+1, sizeof(char));
  ext_unsigned2string<unsigned int>(n, 10, theArray);
  *this = (const char *)theArray;
  free( theArray );
}


OString::OString(long n) : ext_string()
{
  char * theArray = (char *)calloc(sizeof(long)*3+1, sizeof(char));
  ext_signed2string<long>(n, 10, theArray);
  *this = (const char *)theArray;
  free( theArray );
}


OString::OString(unsigned long n) : ext_string()
{
  char * theArray = (char *)calloc(sizeof(unsigned long)*3+1, sizeof(char));
  ext_unsigned2string<unsigned long>(n, 10, theArray);
  *this = (const char *)theArray;
  free( theArray );
}


OString::OString(PInt64 n) : ext_string()
{
  char * theArray = (char *)calloc(sizeof(PInt64)*3+1, sizeof(char));
  ext_unsigned2string<unsigned long>(n, 10, theArray);
  *this = (const char *)theArray;
  free( theArray );
}


OString::OString(PUInt64 n) : ext_string()
{
  char * theArray = (char *)calloc(sizeof(PUInt64)*3+1, sizeof(char));
  ext_unsigned2string<PUInt64>(n, 10, theArray);
  *this = (const char *)theArray;
  free( theArray );
}

template <class InputIterator>
OString::OString(
  InputIterator first, 
  InputIterator last
) : ext_string(first, last) 
{ 
}

template < class T >
OString OString::ToString(const T &arg)
{
	ostringstream	out;

	out << arg;

	return(out.str());
}

OString OString::Trim()const
{
  const char * lpos = c_str();
  const char * theArray = c_str();
  while (isspace(*lpos & 0xff))
    lpos++;
  if (*lpos == '\0')
    return OString();

  const char * rpos = theArray+length()-1;
  if (!isspace(*rpos & 0xff))
    return OString(lpos);

  while (isspace(*rpos & 0xff))
    rpos--;

  return OString(lpos, rpos - lpos + 1);
}

OString OString::ToLower()const 
{
//#ifdef WIN32
  const OString * me = const_cast<OString *>(this);
  char * newPtr = (char*)malloc(length()+1);
  memcpy(newPtr,me->c_str(),length()+1);
  
  for (char *cpos = newPtr; *cpos != '\0'; cpos++) {
    if (isupper(*cpos & 0xff))
      *cpos = (char)::tolower(*cpos & 0xff);
  }
  OString newStr = (const char *)newPtr;

  free( newPtr );
  return newStr;
//#else
//return const_cast<OString*>(this)->tolower();
//#endif

}

OString OString::ToUpper()const 
{
//#ifdef WIN32
  char * newPtr = (char*)malloc(length()+1);
  memcpy(newPtr,c_str(),length()+1);

  for (char *cpos = newPtr; *cpos != '\0'; cpos++) {
    if (islower(*cpos & 0xff))
      *cpos = (char)::toupper(*cpos & 0xff);
  }

  OString newStr = (const char *)newPtr;
  free( newPtr );
  return newStr;
//#else
//return const_cast<OString*>(this)->toupper();
//#endif
}

PINDEX OString::Tokenise(
  vector<OString> & tokens,
  const OString & delimiters,
  BOOL includeBlanks
) const
{
  // Skip delimiters at beginning.
  PINDEX lastPos = find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
  PINDEX pos     = find_first_of(delimiters, lastPos);

  while (((int)string::npos) != pos || ((int)string::npos) != lastPos)
  {
        // Found a token, add it to the vector.
    string token = substr(lastPos, pos - lastPos);
    bool notOk = token.length() == 0 && !includeBlanks;
    
    if( !notOk )
      tokens.push_back( token );

        // Skip delimiters.  Note the "not_of"
    lastPos = find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
    pos = find_first_of(delimiters, lastPos);
  }
  
  if( tokens.size() == 0 )
    tokens.push_back( *this );
  
  return tokens.size();

}


PINDEX OString::GetSize()const
{
  return const_cast<OString*>(this)->length() + 1;
}

PINDEX OString::GetLength()const
{
  return const_cast<OString*>(this)->length();
}

OString OString::Left(
  PINDEX len
)const
{
  if (len <= 0)
    return "";

  if (len >= GetLength())
    return *this;

  return OString(c_str(), len);
}

OString OString::Right(
  PINDEX len
)const
{
  if (len <= 0)
    return "";

  PINDEX srclen = GetLength();
  if (len >= srclen)
    return *this;

  return OString(c_str()+srclen-len, len);
}

OString OString::Mid(
  PINDEX offset
)const
{
  return substr(offset);
}

OString OString::Mid(
  PINDEX offset,
  PINDEX len
)const
{
  return substr(offset, len);
}

long OString::AsInteger(unsigned base) const
{
  return strtol(c_str(), NULL, base);
}


unsigned long OString::AsUnsigned(unsigned base) const
{
  return strtoul(c_str(), NULL, base);
}

OString::Comparison OString::InternalCompare(PINDEX offset, char c) const
{
  const char * theArray = c_str();
  if (offset < 0)
    return LessThan;
  const int ch = theArray[offset] & 0xff;
  if (ch < (c & 0xff))
    return LessThan;
  if (ch > (c & 0xff))
    return GreaterThan;
  return EqualTo;
}


OString::Comparison OString::InternalCompare(
                         PINDEX offset, PINDEX length, const char * cstr) const
{
  const char * theArray = c_str();

  if (offset < 0 || length < 0)
    return LessThan;

  if (offset == 0 && theArray == cstr)
    return EqualTo;

  if (offset < 0 || cstr == NULL)
    return IsEmpty() ? EqualTo : LessThan;

  int retval;
  if (length == P_MAX_INDEX)
    retval = strcmp(theArray+offset, cstr);
  else
    retval = strncmp(theArray+offset, cstr, length);

  if (retval < 0)
    return LessThan;

  if (retval > 0)
    return GreaterThan;

  return EqualTo;
}

PINDEX OString::Find(
  const char * cstr,    ///< C string to search for in string.
  PINDEX offset     ///< Offset into string to begin search.
) const
{
  size_type ret = find( cstr, offset );
  if( ret == string::npos )
    return P_MAX_INDEX;

  return ret;

}

PINDEX OString::FindLast(
  const char * cstr    ///< C string to search for in string.
) const
{
  size_type ret = rfind( cstr );
  if( ret == string::npos )
    return P_MAX_INDEX;
  return ret;
}

PINDEX OString::FindOneOf(const char * cset, PINDEX offset) const
{
  /*
  if (cset == NULL || *cset == '\0' || offset < 0)
    return P_MAX_INDEX;

  PINDEX len = GetLength();
  while (offset < len) {
    const char * p = cset;
    while (*p != '\0') {
      const char * theArray = c_str(); 
      const int ch = theArray[offset] & 0xff;
      if (ch == (*p & 0xff))
        return offset;
      p++;
    }
    offset++;
  }
  return P_MAX_INDEX;
  */
  size_type ret = find_first_of( cset, offset );
  if( ret == string::npos )
    return P_MAX_INDEX;
  return ret;
}


OString OString::Empty()
{
  return OString();
}

bool OString::IsEmpty()const
{
  return empty();
}

BOOL OString::MakeUnique()
{
   return TRUE;
}

void OString::Replace(
  const OString & target,
  const OString & subs,
  BOOL dummy
)
{
  ext_string::replace( target, subs );
}

const  char * OString::GetPointer( int dummy )
{
  dummy = 0;
  return  c_str();
}

OString & OString::operator=( char c )
{
  iterator i = begin();
  *i = c;
  return *this;
}

BOOL OString::operator *= (
  const OString & ostring
)const
{
  return ToLower() == ostring.ToLower();
}

OString * OString::Clone()const
{
  return new OString( *this );
}



//---------------------------------------------------------


OStringArray::OStringArray()
{
}

OStringArray::~OStringArray()
{
}

void OStringArray::AppendString(
  const OString & ostring
)
{
  push_back( ostring );
}

void OStringArray::RemoveAt( PINDEX offset )
{
  iterator iter = begin() + offset;;
  erase( iter );
}

void OStringArray::InsertAt( PINDEX offset,
  const OString & ostring
)
{
  iterator iter = begin() + offset;;
  insert( iter, ostring );
}

PINDEX OStringArray::GetSize()const
{
  return const_cast<OStringArray *>(this)->size();
}

OStringArray * OStringArray::Clone()const
{
  OStringArray * clone = new OStringArray();
  *clone = *this;
  return clone;
}

BOOL OStringArray::IsEmpty()const
{
  return const_cast<OStringArray*>(this)->size() == 0;
}




//---------------------------------------------------------

OStringArray OString::Tokenise(
const OString & delimiters,
BOOL includeBlanks
) const
{
  OStringArray tokens;
  // Skip delimiters at beginning.
  PINDEX lastPos = find_first_not_of(delimiters, 0);
  // Find first "non-delimiter".
  PINDEX pos     = find_first_of(delimiters, lastPos);

  while (((int)string::npos) != pos || ((int)string::npos) != lastPos)
  {
    // Found a token, add it to the vector.
    string token = substr(lastPos, pos - lastPos);
    bool notOk = token.length() == 0 && !includeBlanks;

    if( !notOk )
      tokens.push_back( token );

    // Skip delimiters.  Note the "not_of"
    lastPos = find_first_not_of(delimiters, pos);
    // Find next "non-delimiter"
    pos = find_first_of(delimiters, lastPos);
  }

  if( tokens.size() == 0 )
    tokens.push_back( *this );

  return tokens;

}

OStringArray OString::Lines() const
{
  OStringArray lines;

  if (IsEmpty())
    return lines;

  const char * theArray = c_str();

  //PINDEX line = 0;
  PINDEX p1 = 0;
  PINDEX p2;

  while ((p2 = FindOneOf("\r\n", p1)) != P_MAX_INDEX) {
    lines.push_back( substr( p1, p2 -p1 ) );

    p1 = p2 + 1;
    if (theArray[p2] == '\r' && theArray[p1] == '\n') // CR LF pair
      p1++;
  }

  if (p1 < GetLength())
    lines.push_back( substr( p1, P_MAX_INDEX ) );

  return lines;
}

PINDEX ABSINDEX(PINDEX idx) { return (idx < 0 ? -idx : idx)&P_MAX_INDEX; }

PINDEX OString::HashFunction() const
{
  // Hash function from "Data Structures and Algorithm Analysis in C++" by
  // Mark Allen Weiss, with limit of only executing over first 8 characters to
  // increase speed when dealing with large strings.
  const char * theArray = c_str();
  PINDEX hash = 0;
  for (PINDEX i = 0; i < 8 && theArray[i] != 0; i++)
    hash = (hash << 5) ^ ::tolower(theArray[i] & 0xff) ^ hash;
  return ABSINDEX(hash)%127;
}