/*
 *
 * SIPTransportManager.cxx
 *
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: SIPTransportManager.cxx,v $
 * Revision 1.54  2007/07/15 12:33:22  joegenbaclor
 * Corrected bug in calling OnIncomingSDPOffer where it is called for the second leg instead of the first leg call session.  This is due to the change in placement of the method form receipt of the invite to SetupOutbound
 *
 * Revision 1.53  2007/07/14 09:49:48  joegenbaclor
 * Corrected compile errors in linux
 *
 * Revision 1.52  2007/07/14 07:50:33  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.51  2007/07/10 02:57:40  joegenbaclor
 * Added sanity checks for loopback after calling ExternalRouteTableAnalyzer->OnGetRouteTableEntry
 *
 * Revision 1.50  2007/07/09 03:25:25  joegenbaclor
 * Added resource counter class
 *
 * Revision 1.49  2007/07/07 15:19:14  joegenbaclor
 * Added iproute2 support for linux
 *
 * Revision 1.48  2007/06/14 17:08:52  joegenbaclor
 * Added DNS Caching
 *
 * Revision 1.47  2007/05/14 05:49:33  joegenbaclor
 * Added DNS Failover.
 * Loweded frequency of Timer B and F from 64 to 20
 *
 * Revision 1.46  2007/05/08 12:49:28  joegenbaclor
 * Corrected memory leaks in SIPMessage.
 *
 * Revision 1.45  2007/05/07 16:15:01  joegenbaclor
 * Fixed memory leaks in Garbage Collection
 *
 * Revision 1.44  2007/04/20 06:36:50  joegenbaclor
 * Completed initial work on local REFER
 *
 * Revision 1.43  2007/04/10 06:13:08  joegenbaclor
 * Added new Encryption Mode
 *
 * Revision 1.42  2007/03/28 09:09:24  joegenbaclor
 * Logging enhancements
 *
 * Revision 1.41  2007/03/21 17:37:42  joegenbaclor
 * Added new version XORHash which returns a string to be sued for SIPMessages.
 *
 * Revision 1.40  2007/03/21 05:30:44  joegenbaclor
 * Logging aesthetics
 *
 * Revision 1.39  2007/03/19 02:17:26  joegenbaclor
 * CPU Throttling disabled by default
 *
 * Revision 1.38  2007/03/05 09:29:26  joegenbaclor
 * Some thread optimizations
 *
 * Revision 1.37  2007/03/03 14:51:03  joegenbaclor
 * Added SysInfo
 *
 * Revision 1.36  2007/03/02 09:15:59  joegenbaclor
 * Removed Linux CPU monitor code temporarily
 *
 * Revision 1.35  2007/03/02 07:30:53  joegenbaclor
 * Corrected compile errors from last commit
 *
 * Revision 1.34  2007/03/02 06:22:46  joegenbaclor
 * Corrected some bug in GC.
 * Added CPU Throttle mechanism during heavy load to avoid denial of service
 * Made sure lists are Made unique in SIPHeade clone function
 *
 * Revision 1.33  2007/03/01 10:18:02  joegenbaclor
 * - Added CPU load monitoring
 * - Added Garbage Collection flag to PObject to avoid stale references
 *
 * Revision 1.32  2007/02/28 10:29:57  joegenbaclor
 * EnqueueEvent now calls ProcessOutbound directly
 *
 * Revision 1.31  2007/02/27 23:30:17  joegenbaclor
 * Made sure contact address in 200 OK use the same port used by the INVITE
 * Corrected minor typos from last commit
 *
 * Revision 1.30  2007/02/27 08:18:33  joegenbaclor
 * Introduces new stages for inbound and outbound processing
 *
 * Revision 1.29  2007/02/27 02:11:30  joegenbaclor
 * *** empty log message ***
 *
 * Revision 1.28  2007/02/25 16:13:09  joegenbaclor
 * Route Table and Interface Table variables are now static
 *
 * Revision 1.27  2007/02/22 14:40:38  joegenbaclor
 * Completed multi listener support
 *
 * Revision 1.26  2007/02/04 15:04:10  joegenbaclor
 * Completed transaction layer TCP support
 *
 * Revision 1.25  2007/01/29 08:35:58  joegenbaclor
 * More work on TCP Transport
 *
 * Revision 1.24  2007/01/27 23:06:51  joegenbaclor
 * More work on TCP support
 *
 * Revision 1.23  2007/01/10 09:39:15  joegenbaclor
 * minor bug fix related to last commit
 *
 * Revision 1.22  2007/01/10 09:03:56  joegenbaclor
 * Modified GetDefaultInterfaceAddress() to return the mask entry if size is one
 *
 * Revision 1.21  2007/01/06 07:30:33  joegenbaclor
 * Added mask to filter GetDefaultInterfaceAddress
 *
 * Revision 1.20  2006/12/22 07:37:21  joegenbaclor
 * Introduced separate funtion call  for starting of transport thread.  This is necessary
 *  to make sure that the backdoor listener is already initialized prior to calling transport
 *  reads
 *
 * Revision 1.19  2006/12/06 06:40:56  joegenbaclor
 * STUN related fixes
 *
 * Revision 1.18  2006/11/30 08:19:38  joegenbaclor
 * Relicensed opensipstack to MPL/GPL/LGPL triple license scheme
 *
 * Revision 1.17  2006/11/21 06:55:19  joegenbaclor
 * Added ability to add/append additional listeners.
 *
 * Revision 1.16  2006/11/16 01:35:59  rcolobong
 * Fixed bug in GetDefaultInterfaceAddress during failure to retrieve route table for a private network
 *
 * Revision 1.15  2006/11/09 12:48:19  joegenbaclor
 * Fixed bug in GetDefaultInterfaceAddress() caused by none retrieval of the route table
 *
 * Revision 1.14  2006/08/28 07:18:35  joegenbaclor
 * Bulk Commit: Logging overhaul and Call Session callback changes to reflect SIPMessage eg:  OnAlerting(), OnProgress(), etc
 *
 * Revision 1.13  2006/04/20 14:32:28  joegenbaclor
 * Fixed early media bug where soucd channels in OPAL ay not be opened in time for ringback to start streaming.
 *
 * Revision 1.12  2006/04/05 00:26:04  joegenbaclor
 * 1.  Completed support for STUN NAT traversal method and OPAL style address translation
 * 2.  More bug fixes to ProxySession
 * 3.  Fixed bug in messages where zero size list headers may be interpreted by the parser as a valid header
 * 4.  Started work on OPAL outbound
 *
 * Revision 1.11  2006/03/21 14:11:51  joegenbaclor
 * 1.  Added rport support for NAT Traversal
 * 2.  Added NAT Keep alive mechanism in SIPSession
 * 3.  Some Improvements in various classes affecting NAT like Via, SIPTransport etc.
 *
 * Revision 1.10  2006/03/21 09:03:57  joegenbaclor
 * Introduced Terminate() method in SIPTransportManager to allow
 * SIPStack to terminate Transport threads prior to the deletion of the manager
 *
 * Revision 1.9  2006/03/16 14:44:18  joegenbaclor
 * 1.  Implemented Clone function in all SDPHeader decendant.
 * 2.  Fixed bug in mime param where "=" characters in the param value may cause the parser to barf.
 * 3.  Implemented Proxy Authentication in CallSession
 * 4.  More work on Opal OSS Classes
 * 5.  Changed UDP Transport write behavior to send outbound message in all interfaces if GetSocket() returns NULL.
 *      Old behavior was to send the packet using the default interface.
 *
 * Revision 1.8  2006/03/14 03:53:53  joegenbaclor
 * Initial source upload for BETA Release Candidate
 *
 *
 *
 */

#include "SIPTransportManager.h"
#include "SIPUDPTransport.h"
#include "SIPTCPTransport.h"
#include "SysInfo.h"
#include "Encryption.h"

using namespace SIPTransports;
using namespace NAT;

#define new PNEW

PIPSocket::RouteTable SIPTransportManager::m_RouteTable;
BOOL SIPTransportManager::m_HasRouteTable = FALSE;
PIPSocket::InterfaceTable SIPTransportManager::m_IfTable;
BOOL SIPTransportManager::m_HasIfTable = FALSE;
SIPTransportManager::ExternalRouteTableAnalyzer * SIPTransportManager::m_ExternalRouteTableAnalyzer = NULL;

#if ENABLE_CPU_YIELD

class CPUMonitor : public PThread
{
  PCLASSINFO( CPUMonitor, PThread );
public:
  CPUMonitor() : PThread( 1024, PThread::NoAutoDeleteThread )
  {
    m_CanReject = TRUE;
    Resume();
  }

  ~CPUMonitor()
  { 
    m_ExitSync.Signal();
    WaitForTermination();
  }

  void Main()
  {
    while( !m_ExitSync.Wait( 500 ) )
    {
      unsigned long load;
      if( Tools::SysInfo::GetCPUUsage( load ) )
      {
        m_CPULoad = load;
        m_CanReject = !m_CanReject;
      }
    }
  }

  BOOL IsResourceAvailable()
  {
    PWaitAndSignal lock( m_Resource );

    static int rejectBase = 0;
    rejectBase++;
    BOOL accept = TRUE;

    if( m_CanReject )
    {
      if( m_CPULoad > 50 && m_CPULoad < 60 && rejectBase == 10 )
      {
        accept = FALSE;
      }else if( m_CPULoad > 80 && m_CPULoad < 90 )
      {
        if( rejectBase == 5 || rejectBase == 10 )
          accept = FALSE;
      }else if( m_CPULoad >= 90 && m_CPULoad < 100 )
      {
        if( rejectBase == 3 || rejectBase == 6 || rejectBase == 9 )
          accept = FALSE;
      }else if( m_CPULoad >= 100 )
      {
        accept = FALSE;
      }
    }

    if( rejectBase >= 10 )
      rejectBase = 0;

    return accept;
  }

  PSyncPoint m_ExitSync;
  PAtomicInteger m_CPULoad;
  PMutex m_Resource;
  BOOL m_CanReject;
};

#endif

SIPTransportManager::SIPTransportManager(
  PNotifier eventNotifier
)
{
  m_EventNotifier = eventNotifier;
  m_STUNClient = NULL;
  m_UDPTransport = NULL;
  m_TCPTransport = NULL;
  m_TLSTransport = NULL;
  m_IsTerminated = FALSE;
  m_HasRouteTable = PIPSocket::GetRouteTable(m_RouteTable);
  m_HasIfTable = PIPSocket::GetInterfaceTable( m_IfTable );
  m_InboundProcessor = 
        new EQ::EventQueue( PCREATE_NOTIFIER( OnProcessInbound ), 
        1, 1024 * 2, P_MAX_INDEX, "Transport (INBOUND)", PThread::HighestPriority);
  m_OutboundProcessor = 
        new EQ::EventQueue( PCREATE_NOTIFIER( OnProcessOutbound ), 
        1, 1024 * 2,  P_MAX_INDEX, "Transport (OUTBOUND)", PThread::HighestPriority);
#if ENABLE_CPU_YIELD
  m_CPUMonitorThread = new  CPUMonitor();
#else
  m_CPUMonitorThread = NULL;
#endif
  
  SIPTransport::Initialize();
  SIPSrvRecord::InitializeCacheManagers();
}

SIPTransportManager::~SIPTransportManager()
{
  Terminate();
  if( m_InboundProcessor != NULL )
    delete m_InboundProcessor;

  if( m_OutboundProcessor != NULL )
    delete m_OutboundProcessor;

  if( m_CPUMonitorThread != NULL )
    delete m_CPUMonitorThread;

  SIPSrvRecord::DestroyCacheManagers();
}

BOOL SIPTransportManager::Terminate()
{
  if( m_IsTerminated )
    return TRUE;

  m_IsTerminated = TRUE;

  if( m_UDPTransport != NULL )
  {
    m_UDPTransport->CleanUp();
    delete m_UDPTransport;
  }

  if( m_TCPTransport != NULL )
  {
    m_TCPTransport->CleanUp();
    delete m_TCPTransport;
  }

  if( m_TLSTransport != NULL )
  {
    m_TLSTransport->CleanUp();
    delete m_TLSTransport;
  }

  if( m_STUNClient != NULL )
    delete m_STUNClient;

  SIPTransport::Deinitialize();

  return TRUE;
}

BOOL SIPTransportManager::AppendTransport(
  SIPTransport::Protocol protocol,
  const OString & iface,
  WORD port
)
{
  return AddTransport( protocol, iface, port, TRUE );
}

BOOL SIPTransportManager::AppendTransport(
  SIPTransport::Protocol protocol,
  const PIPSocket::Address & iface,
  WORD port
)
{
  return AddTransport( protocol, iface, port, TRUE );
}

BOOL SIPTransportManager::AddTransport(
  SIPTransport::Protocol protocol,
  const PIPSocket::Address & _iface,
  WORD port,
  BOOL append
)
{
  OString iface = "*";
  if( _iface != 0 )
  {
    iface = (const char *)_iface.AsString();
  }

  BOOL ret =  AddTransport( protocol, iface, port, append );

  return ret;
}

BOOL SIPTransportManager::AddTransport(
  SIPTransport::Protocol protocol,
  const OString & iface,
  WORD port,
  BOOL append
)
{
  SIPTransport * transport = NULL;

  if( !append )
  {
    transport = OnCreateTransport( protocol );
    if( transport == NULL )
    {
      LOG_IF_DEBUG( LogError(), "*** TRANSPORT ERROR *** " << "Unable to create transport object on " << iface << ":" << port );
      return FALSE;
    }

    if( protocol == SIPTransport::UDP && m_UDPTransport == NULL )
    {
      LOG_IF_DEBUG( LogDetail(), "Starting UDP Transport on " << iface << ":" << port );
      m_UDPTransport = transport;
    }else if( protocol == SIPTransport::TCP && m_TCPTransport == NULL )
    {
       LOG_IF_DEBUG( LogDetail(), "Starting TCP Transport on " << iface << ":" << port );
      m_TCPTransport = transport;
    }else if( protocol == SIPTransport::TLS && m_TLSTransport == NULL )
    {
       LOG_IF_DEBUG( LogDetail(), "Starting TLS Transport on " << iface << ":" << port );
      m_TLSTransport = transport;
    }else
    {
      LOG_IF_DEBUG( LogError(), "*** TRANSPORT ERROR *** " << "Starting Transport on " << iface << ":" << port );
      transport->CleanUp();
      delete transport;
      return FALSE;
    }

    
  }else
  {
    if( protocol == SIPTransport::UDP && m_UDPTransport != NULL )
    {
      transport = m_UDPTransport;
      LOG_IF_DEBUG( LogDetail(), "Starting UDP Transport on " << iface << ":" << port );

    }else if( protocol == SIPTransport::TCP && m_TCPTransport != NULL )
    {
      transport = m_TCPTransport;
      LOG_IF_DEBUG( LogDetail(), "Starting TCP Transport on " << iface << ":" << port );
    }else if( protocol == SIPTransport::TLS && m_TLSTransport != NULL )
    {
      transport = m_TLSTransport;
      LOG_IF_DEBUG( LogDetail(), "Starting TLS Transport on " << iface << ":" << port );
    }else
    {
      return FALSE;
    }
  }

  BOOL ok = transport->Listen( iface, port, append );

  if( ok )
  {
    LOG_IF_DEBUG( LogDetail(), "*** TRANSPORT STARTED *** " << " Listener: " << iface << ":" << port );
  }else
  {
    LOG_IF_DEBUG( LogError(), "*** TRANSPORT FAILED *** " << " Listener: " << iface << ":" << port );
  }

  return ok;
}

BOOL SIPTransportManager::HasListener()const
{
  /// we only support UDP for now.
  if( m_UDPTransport == NULL )
    return FALSE;

  return m_UDPTransport->IsListening();
}

SIPTransport * SIPTransportManager::OnCreateTransport(
  SIPTransport::Protocol protocol
)
{
  if( protocol == SIPTransport::UDP )
    return new SIPUDPTransport( this );
  else if( protocol == SIPTransport::TCP )
    return new SIPTCPTransport( this );

  return NULL;
}



/////////////////////
BOOL SIPTransportManager::EnqueueEvent(
  SIPTransportEvent * event
)
{

  SIPMessage * msg = (SIPMessage*)event->GetEvent();
  ProcessOutbound( new SIPMessage( *msg ) );
  delete event;
  return TRUE;

#if 0
  //PAssertNULL( event );
  SIPTransport * transport = NULL;
  if( event->m_Type == SIPTransportEvent::UDPPacketSend && m_UDPTransport != NULL )
  {
    transport = m_UDPTransport;
  }else if( event->m_Type ==  SIPTransportEvent::TCPPacketSend && m_TCPTransport != NULL )
  {
    transport = m_TCPTransport;
  }else if( event->m_Type ==  SIPTransportEvent::TLSPacketSend && m_TLSTransport != NULL )
  {
    transport = m_TLSTransport;
  }else if( event->m_Type == SIPTransportEvent::PacketSend )
  {
    if( !msg->IsKeepAlive() )
    {
      SIPSrvRecord srvRec;
      srvRec = *msg;

      SIPTransport::Protocol serverProto = DetermineServerProtocol( event );

      if( srvRec.GetSize( "UDP" ) > 0 && serverProto == SIPTransport::UDP )
        transport = m_UDPTransport;
      else if( srvRec.GetSize( "TCP" ) > 0 && serverProto == SIPTransport::TCP )
        transport = m_TCPTransport;
      else if( srvRec.GetSize( "TLS" ) > 0 && serverProto == SIPTransport::TLS )
        transport = m_TLSTransport;
      else
        transport = m_UDPTransport;
    }else
    {
      transport = m_UDPTransport;
    }
  }else
  {
    delete event;
    return FALSE;
  }

  if( transport != NULL )
    transport->Enqueue( *msg );

  delete event;

  return TRUE;
#endif

  
}

#ifdef _WIN32
#pragma warning(disable:4311)
#endif

void SIPTransportManager::OnTransportEvent(
  SIPTransportEvent * event
)
{
  m_EventNotifier( *this, (INT)event );
}

SIPTransport::Protocol SIPTransportManager::DetermineServerProtocol(
  SIPTransportEvent * event
)
{
  SIPMessage * msg = (SIPMessage *)event->GetEvent();
  OString transport;

  if( msg->IsRequest() )
  {
    /// check the URI protocol
    StartLine startLine;
    msg->GetStartLine( startLine );
    const SIPURI & requestURI = startLine.GetRequestLine()->GetRequestURI();
    requestURI.GetParameter( "transport", transport );
  }else
  {
    /// check the via protocol
    if( msg->GetViaSize() > 0 )
    {
      Via via;
      msg->GetViaAt( 0, via );
      transport = via.GetProtocol();
    }
  }

  if( transport *= "udp" )
    return SIPTransport::UDP;
  else if( transport *= "tcp" )
    return SIPTransport::TCP;
  else if( transport *= "tls" )
    return SIPTransport::TLS;
  else
    return SIPTransport::Unknown;
}




PIPSocket::Address SIPTransportManager::GetDefaultInterfaceAddress( 
  BOOL includeLoopBack,
  OStringArray * mask 
)
{
  PIPSocket::Address gateway;
  if( !PIPSocket::GetGatewayAddress( gateway ) )
  {
    PTRACE( 1, "Unable to retrieve gateway address" );
    return 0;
  }

  return GetDefaultInterfaceAddress( gateway, includeLoopBack, mask );
}

static BOOL IsAddressInMask( PIPSocket::Address & addr, OStringArray & mask )
{
  OString address = addr.AsString();
  int mSize = mask.GetSize();
  for( PINDEX i = 0; i < mSize; i++ )
    if( address == mask[i] )
      return TRUE;

  return FALSE;
}



PIPSocket::Address SIPTransportManager::GetDefaultInterfaceAddress(
  const PIPSocket::Address & destination,
  BOOL includeLoopBack,
  OStringArray * mask 
)
{
  if( destination.IsLoopback() )
    return destination;

  if( mask != NULL  && mask->GetSize() == 1 )
  {
    return PIPSocket::Address( (*mask)[0] );
  }
  
  if( SIPTransportManager::m_ExternalRouteTableAnalyzer != NULL )
  {
    PIPSocket::Address route, source;
    OString device;
    if( SIPTransportManager::m_ExternalRouteTableAnalyzer->OnGetRouteTableEntry( destination, source, route, device ) )
    {
      if( source.IsLoopback() && !includeLoopBack )
      {
        /// got a loopback address.  Do nothing
      }else
      {
        if( mask != NULL && IsAddressInMask( source, *mask ) )
          return source;
        else if( mask == NULL )
          return source;
      }
    }
  }

  if( !m_HasIfTable )
  {
    PTRACE( 1, "Unable to retrieve interface table" );
    return 0;
  }

  int ifSize = m_IfTable.GetSize();

  if ( !m_HasRouteTable )
  {
    PTRACE( 1, "Unable to retrieve route table" );

    ///lets use the interface table instead and guess the best interface
    PINDEX i = 0;
    /// get the first public interface and set it as the default address
    if( !SIPTransport::IsPrivateNetwork( destination ) )
    {
      
      for( i = 0; i < ifSize; i++ )
      {
        PIPSocket::Address addr = m_IfTable[i].GetAddress();
        if( !addr.IsLoopback() && !SIPTransport::IsPrivateNetwork( addr ) )
        {
          if( mask != NULL && IsAddressInMask( addr, *mask ) )
            return addr;
        }
      }

      /// mask comparison failed.  we are desperate just use the first interface
      for( i = 0; i < ifSize; i++ )
      {
        PIPSocket::Address addr = m_IfTable[i].GetAddress();
        if( !addr.IsLoopback() && !SIPTransport::IsPrivateNetwork( addr ) )
          return addr;
      }
    }else
    {
      /// still we dont have a default address, lets use the first none loopback address
      for( i = 0; i < ifSize; i++ )
      {
        PIPSocket::Address addr = m_IfTable[i].GetAddress();
        if( !addr.IsLoopback() && SIPTransport::IsPrivateNetwork( addr ) )
          if( mask != NULL && IsAddressInMask( addr, *mask ) )
            return addr;
      }

      /// mask comparison failed.  we are desperate just use the first private interface
      for( i = 0; i < ifSize; i++ )
      {
        PIPSocket::Address addr = m_IfTable[i].GetAddress();
        if( !addr.IsLoopback() && SIPTransport::IsPrivateNetwork( addr ) )
          return addr;
      }

      /// no private interface available, just use the first public interface
      for( i = 0; i < ifSize; i++ )
      {
        PIPSocket::Address addr = m_IfTable[i].GetAddress();
        if( !addr.IsLoopback() && !SIPTransport::IsPrivateNetwork( addr ) )
          return addr;
      }
    }

    return 0;
  }

  // Ok we have a route table
  PIPSocket::Address localAddress;
  int rtSize = m_RouteTable.GetSize();

  for( PINDEX i = 0; i < rtSize; i++ )
  {
    PIPSocket::RouteEntry & entry = m_RouteTable[i];

    if( entry.GetNetMask() == INADDR_ANY  )
      continue;

    
    if( m_RouteTable[i].GetDestination().IsLoopback() && !includeLoopBack )
      continue;

    PIPSocket::Address netMask = entry.GetNetMask();
    if( entry.GetDestination() == destination ||
      ((destination & netMask) == entry.GetNetwork() ) )
    {
      localAddress = m_RouteTable[i].GetDestination();


      if( localAddress == 0 )
      {
        for( PINDEX j = 0; j < ifSize; j++ )
        {
          if( m_IfTable[j].GetName() == entry.GetInterface() )
          {
            localAddress = m_IfTable[j].GetAddress();
          }
        }
      }

      if( localAddress.IsLoopback() && !includeLoopBack )
        continue;

      if( mask != NULL && !IsAddressInMask( localAddress, *mask ) )
        continue;

      return localAddress;
    }
  }

  if( localAddress.IsLoopback() && !includeLoopBack )
    localAddress = GetDefaultInterfaceAddress( FALSE );

  if( mask != NULL && !IsAddressInMask( localAddress, *mask ) )
    return 0;

  return localAddress;
}


SIPSTUNClient * SIPTransportManager::GetSTUN(const PIPSocket::Address & ip) const
{
  if (ip.IsValid() && SIPTransport::IsLocalAddress(ip))
    return NULL;

  return m_STUNClient;
}


SIPSTUNClient::NatTypes SIPTransportManager::SetSTUNServer(
  const OString & server,
  WORD udpPortBase,
  WORD udpPortMax,
  WORD rtpPortBase,
  WORD rtpPortMax
)
{
  if( m_STUNClient != NULL )
    delete m_STUNClient;

  if (server.IsEmpty()) {
    m_STUNClient = NULL;
    return SIPSTUNClient::UnknownNat;
  }

    m_STUNClient = new SIPSTUNClient(server,
                           udpPortBase, udpPortMax,
                           rtpPortBase, rtpPortMax );
  SIPSTUNClient::NatTypes type = m_STUNClient->GetNatType();
  if (type != SIPSTUNClient::BlockedNat && type != SIPSTUNClient::OpenNat)
  {
    m_STUNClient->GetExternalAddress(m_TranslationAddress);
  }else
  {
    delete m_STUNClient;
    m_STUNClient = NULL;
  }

  LOG_IF_DEBUG( LogInfo(), "STUN server \"" << server << "\" replies " << type << ", external IP " << m_TranslationAddress);

  OStringStream strm;
  strm << "STUN server \"" << server << "\" replies " << type << ", external IP " << m_TranslationAddress << "\r\n";
  SIPTransport::NotifyRead( strm.str() );

  return type;
}

BOOL SIPTransportManager::TranslateIPAddress(
  PIPSocket::Address & localAddress,
  const PIPSocket::Address & remoteAddress
 )
{
  if (m_TranslationAddress.IsLoopback())
    return FALSE; // Have nothing to translate it to

  if (!m_TranslationAddress.IsValid())
    return FALSE; // Have nothing to translate it to

  if (!SIPTransport::IsLocalAddress(localAddress))
    return FALSE; // Is already translated

  if (SIPTransport::IsLocalAddress(remoteAddress))
    return FALSE; // Does not need to be translated

  // Tranlsate it!
  localAddress = m_TranslationAddress;
  return TRUE;
}

BOOL SIPTransportManager::GetListenerAddress(
  SIPTransport::Protocol proto,
  const PIPSocket::Address & target,
  PIPSocket::Address & listenerAddress,
  WORD & listenerPort
)
{
  BOOL ok = FALSE;

  if( proto == SIPTransport::UDP && m_UDPTransport != NULL )
  {
    ok = m_UDPTransport->GetListenerAddress( target, listenerAddress, listenerPort );
  }else if( proto == SIPTransport::TCP && m_TCPTransport != NULL )
  {
    ok = FALSE;
  }else if( proto == SIPTransport::TLS && m_TLSTransport != NULL )
  {
    ok = FALSE;
  }

  TranslateIPAddress( listenerAddress, listenerPort );

  return ok;
}

BOOL SIPTransportManager::IsLocalAddressAndPort(
  const SIPURI & uri
)const
{
  PIPSocket::Address addr;
  WORD port;
  if( !SIPTransport::Resolve( uri, addr, port ) )
    return FALSE;

  return IsLocalAddressAndPort( addr, port );
}

BOOL SIPTransportManager::IsLocalAddressAndPort(
  const PIPSocket::Address & addr,
  WORD port
)const
{
  if( port == 0 )
    port = 5060;

  return m_UDPTransport->IsLocalAddressAndPort( addr, port );
}

BOOL SIPTransportManager::StartThreads()
{

  if( m_UDPTransport != NULL )
    if( !m_UDPTransport->StartThreads() )
      return FALSE;

  if( m_TCPTransport != NULL )
    if( !m_TCPTransport->StartThreads() )
      return FALSE;

  if( m_TLSTransport != NULL )
    if( !m_TLSTransport->StartThreads() )
      return FALSE;

  return TRUE;
}

BOOL SIPTransportManager::ProcessPacket(
  SIPPacket * packet
)
{
  if( m_IsTerminated )
  {
    delete packet;
    return FALSE;
  }

  BOOL ok = m_InboundProcessor->EnqueueEvent( packet );
  
  return ok;
}

BOOL SIPTransportManager::ProcessOutbound(
  SIPMessage * packet
)
{
  if( packet == NULL )
    return FALSE;

  if( m_IsTerminated )
  {
    delete packet;
    return FALSE;
  }

  return m_OutboundProcessor->EnqueueEvent( packet );
}

void SIPTransportManager::OnProcessInbound( EQ::EventQueue &, INT _event )
{
  SIPPacket * packet = reinterpret_cast<SIPPacket *>( _event );
  if( packet == NULL )
    return;
  int bytesRead = packet->GetSize();

  if( m_IsTerminated )
  {
    delete packet;
    return;
  }

  if( bytesRead < MIN_SIP_MESSAGE_LENGTH )
  {
    delete packet;
    return;
  }

  if( packet->GetTransportProtocol() == SIPTransport::UDP )
  {
#if 1   //ENABLE_GRANSTREAM_XOR
    PBYTEArray temp(bytesRead);
    ::memcpy( temp.GetPointer(bytesRead), packet->GetPointer(bytesRead), bytesRead );
    temp.SetSize( bytesRead + 1 );
    temp[bytesRead]='\0';
    OString buff = reinterpret_cast<const char*>( temp.GetPointer(temp.GetSize()) );
    buff = ParserTools::LineFeedSanityCheck( buff );
    SIPMessage * msg = new SIPMessage( buff );

    if( msg->GetCSeqMethod().IsEmpty() )
    {
      /// create a clean slate
      delete msg;
      msg = new SIPMessage();

      PBYTEArray in = *packet;
      OString out;
      LOG_IF_DEBUG( LogDetail(), in );

      Encryption::Engine::Decrypt( in, out );
      
      LOG_IF_DEBUG( LogDebugVeryHigh(), out );
      buff = out;
      buff = ParserTools::LineFeedSanityCheck( buff );
      *msg = buff;


      if( !msg->GetCSeqMethod().IsEmpty() )
         msg->SetEncryption( TRUE );

    }
#else
    packet->SetSize( bytesRead + 1 );
    (*packet)[bytesRead]='\0';
    OString buff( (char*)packet->GetPointer() );
    buff = ParserTools::LineFeedSanityCheck( buff );
    msg = buff;
#endif

    if( !msg->IsValid() )
    {
      delete msg;
      delete packet;
      return;
    }

#if ENABLE_CPU_YIELD
    if( msg->IsInvite() && !dynamic_cast<CPUMonitor*>(m_CPUMonitorThread)->IsResourceAvailable() )
    {
      /// Throttle INVITE based on CPU availability
      delete msg;
      delete packet;
      return;
    }
#endif
    
    /// Set Receive Address custom Headers for NAT traversal support
    OString addrStr = packet->GetSourceAddress().AsString();
    OString portStr( packet->GetSourcePort() );
    SIPHeader rcvAddr( "RCVADDR", addrStr );
    SIPHeader rcvPort( "RCVPORT", portStr );
    SIPHeader rcvTran( "RCVTRAN", "udp" );
    
    msg->AddInternalHeader( rcvAddr );
    msg->AddInternalHeader( rcvPort );
    msg->AddInternalHeader( rcvTran );
    msg->SetInterfaceAddress( (const char *)packet->GetLocalAddress().AsString() );
    msg->SetInterfacePort( packet->GetLocalPort() );

    LOG_CONTEXT_IF_DEBUG( LogDebug(), msg->GetCallId(), "RCV: enc=" << msg->IsEncrypted() 
      << " " << bytesRead << " Bytes from " << rcvAddr << ":" 
      << rcvPort << ":UDP (" << msg->GetStartLine() << ")" );
    
    OStringStream traceStream;
    traceStream << "RCV: enc=" << msg->IsEncrypted() << " " 
      << bytesRead << " Bytes from " << rcvAddr << ":" << rcvPort 
      << ":UDP (" << msg->GetStartLine() << ")" 
      << "\r\n" << *msg;

    PTRACE( 5, traceStream.str() );

    SIPTransport::NotifyRead( traceStream.str() );

    if( msg->IsInvite() )
    {
      SIPMessage * trying = new SIPMessage();
      msg->CreateResponse( *trying, SIPMessage::Code100_Trying );
      if( msg->IsEncrypted() )
        trying->SetEncryption( TRUE );
      ProcessOutbound( trying );
    }

    OnTransportEvent( 
    new SIPTransportEvent( 
      msg, 
      SIPTransportEvent::UDPPacketArrival 
    ) );

  }else if( packet->GetTransportProtocol() == SIPTransport::TCP )
  {
  }else if( packet->GetTransportProtocol() == SIPTransport::TLS )
  {
  }else
  {
  }


  delete packet;
}

void SIPTransportManager::OnProcessOutbound( EQ::EventQueue &, INT _event )
{
  SIPMessage * packet = reinterpret_cast<SIPMessage *>( _event );
  if( packet == NULL )
    return;

  PIPSocket::Address sendAddr;
  WORD sendPort;

  SIPSrvRecord srvRec;
  srvRec = *packet;
  
  if( srvRec.GetSize( "UDP" ) == 0)
  {
    delete packet;
    return;
  }

  const SIPSrvRecord::HostRecord * rec = srvRec.GetRecordAt( 0, "UDP" );
  sendAddr = rec->m_Address;
  sendPort = rec->m_Port;

  if( packet->IsRequest() )
  {
    Via via = packet->PopTopVia();
    via.SetParameter( "uas-addr", (const char *)sendAddr.AsString() );
    packet->AppendVia( via );
  }

  OString sendPacket;
  OString ifaceAddress = packet->GetInterfaceAddress();
  WORD ifacePort = packet->GetInterfacePort();
  sendPacket = packet->AsString();

  if( !packet->IsKeepAlive() )
  {
    OStringStream traceStream;
    traceStream <<  "SEND: enc=" << packet->IsEncrypted() 
      << " " <<  sendPacket.GetLength() << " Bytes to " 
      << sendAddr << ":" << sendPort << ":UDP (" 
      << packet->GetStartLine() << ")" 
      << " Interface Address=" << ifaceAddress
      << "\r\n" << *packet;

    LOG_CONTEXT_IF_DEBUG( LogDebug(), packet->GetCallId(), traceStream.str() );
    SIPTransport::NotifyWrite( traceStream.str() );
  }else
    sendPacket = "\r\n\r\n";

  BOOL ok;

  if( packet->IsEncrypted() )
  {
    PBYTEArray out;
    packet->GetEncrypted( out );
    ok = m_UDPTransport->WriteBytes( out, sendAddr, sendPort, ifaceAddress, ifacePort );
  }else
  {
    ok = m_UDPTransport->WriteString( sendPacket, sendAddr, sendPort, ifaceAddress, ifacePort );
  }


  BOOL isKeepalive = packet->IsKeepAlive();

  if( isKeepalive )
  {
    delete packet;
    return;
  }
  
  if( ok )
  {
    if( packet->IsInvite() )
    {
      OnTransportEvent( 
      new SIPTransportEvent( 
        packet, 
        SIPTransportEvent::UDPPacketSent 
      ) );
    }else
    {
      delete packet;
    }
  }else
  {
    OnTransportEvent( 
    new SIPTransportEvent( 
      packet, 
      SIPTransportEvent::UDPWriteError
    ) );
  }
  return ;
}

